from typing import cast
from unittest import TestCase

from pyinfra.api import Config, Host, Inventory, State
from pyinfra.api.arguments import AllArguments, pop_global_arguments


class TestOperationKwargs(TestCase):
    def test_get_from_config(self):
        config = Config(SUDO="config-value")
        inventory = Inventory((("somehost",), {}))

        state = State(config=config, inventory=inventory)
        host = inventory.get_host("somehost")

        kwargs, _ = pop_global_arguments(state, host, {})
        assert kwargs.get("_sudo") == "config-value"

    def test_get_from_host(self):
        config = Config(SUDO="config-value")
        inventory = Inventory(([("somehost", {"_sudo": True})], {}))

        state = State(config=config, inventory=inventory)
        host = inventory.get_host("somehost")

        kwargs, _ = pop_global_arguments(state, host, {})
        assert kwargs.get("_sudo") is True

    def test_get_from_state_deploy_kwargs(self):
        config = Config(SUDO="config-value")
        inventory = Inventory(([("somehost", {"_sudo": False})], {}))
        somehost = inventory.get_host("somehost")
        assert isinstance(somehost, Host)

        state = State(config=config, inventory=inventory)
        host = inventory.get_host("somehost")
        somehost.current_deploy_kwargs = cast(AllArguments, {"_sudo": True})

        kwargs, keys = pop_global_arguments(state, host, {})
        assert kwargs.get("_sudo") is True

    def test_get_from_kwargs(self):
        config = Config(SUDO="config-value")
        inventory = Inventory(([("somehost", {"_sudo": False})], {}))
        somehost = inventory.get_host("somehost")
        assert isinstance(somehost, Host)

        state = State(config=config, inventory=inventory)
        somehost.current_deploy_kwargs = cast(
            AllArguments,
            {
                "_sudo": False,
                "_sudo_user": "deploy-kwarg-user",
            },
        )

        kwargs, keys = pop_global_arguments(state, somehost, {"_sudo": True})
        assert kwargs.get("_sudo") is True
        assert kwargs.get("_sudo_user") == "deploy-kwarg-user"
        assert "_sudo" in keys



import os
from pathlib import Path

import requests

from scripts.generate_next_version import get_version_from_changelog

GITHUB_API_TOKEN = os.environ["GITHUB_API_TOKEN"]


def make_github_release():
    version = get_version_from_changelog()
    tag = f"v{version}"

    changelog_path = Path(__file__).parent.parent / "CHANGELOG.md"

    with open(changelog_path, "r", encoding="utf-8") as f:
        changelog_data = f.read()

    changelog_lines = []

    for line in changelog_data.splitlines():
        if line.startswith("# "):
            if line != f"# {tag}":
                break
            continue

        changelog_lines.append(line)

    changelog = "\n".join(changelog_lines).strip()

    response = requests.post(
        "https://api.github.com/repos/pyinfra-dev/pyinfra/releases",
        json={
            "tag_name": tag,
            "body": changelog,
        },
        headers={
            "Authorization": f"token {GITHUB_API_TOKEN}",
        },
    )
    response.raise_for_status()


if __name__ == "__main__":
    make_github_release()



"""
Fetch and install binary updates to FreeBSD.
"""

from __future__ import annotations

from typing_extensions import List, Optional, Union

from pyinfra.api import QuoteString, StringCommand, operation


@operation()
def update(
    force: bool = False,
    basedir: Optional[str] = None,
    workdir: Optional[str] = None,
    conffile: Optional[str] = None,
    jail: Optional[str] = None,
    key: Optional[str] = None,
    currently_running: Optional[str] = None,
    server: Optional[str] = None,
):
    """
    Based on the currently installed world and the configuration options set, fetch
    all available binary updates and install them.

    + force: See ``-F`` in ``freebsd-update(8)``.
    + basedir: See ``-b`` in ``freebsd-update(8)``.
    + workdir: See ``-d`` in ``freebsd-update(8)``.
    + conffile: See ``-f`` in ``freebsd-update(8)``.
    + jail: See ``-j`` in ``freebsd-update(8)``.
    + key: See ``-k`` in ``freebsd-update(8)``.
    + currently_running: See ``--currently-running`` in ``freebsd-update(8)``.
    + server: See ``-s`` in ``freebsd-update(8)``.

    **Example:**

    .. code:: python

        freebsd_update.update()
    """

    args: List[Union[str, "QuoteString"]] = []

    args.extend(["PAGER=cat", "freebsd-update", "--not-running-from-cron"])

    if force:
        args.append("-F")

    if basedir is not None:
        args.extend(["-b", QuoteString(basedir)])

    if workdir is not None:
        args.extend(["-d", QuoteString(workdir)])

    if conffile is not None:
        args.extend(["-f", QuoteString(conffile)])

    if jail is not None:
        args.extend(["-j", QuoteString(jail)])

    if key is not None:
        args.extend(["-k", QuoteString(key)])

    if server is not None:
        args.extend(["-s", QuoteString(server)])

    args.extend(["fetch", "install"])

    yield StringCommand(*args)



import json

from typing_extensions import override

from pyinfra import local, logger
from pyinfra.api.exceptions import InventoryError
from pyinfra.api.util import memoize
from pyinfra.progress import progress_spinner

from .base import BaseConnector


@memoize
def show_warning() -> None:
    logger.warning("The @terraform connector is in beta!")


def _flatten_dict_gen(d, parent_key, sep):
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        yield new_key, v
        if isinstance(v, dict):
            yield from _flatten_dict(v, new_key, sep=sep).items()


def _flatten_dict(d: dict, parent_key: str = "", sep: str = "."):
    return dict(_flatten_dict_gen(d, parent_key, sep))


class TerraformInventoryConnector(BaseConnector):
    """
    Generate one or more SSH hosts from a Terraform output variable. The variable
    must be a list of hostnames or dictionaries.

    Output is fetched from a flattened JSON dictionary output from ``terraform output
    -json``. For example the following object:

    .. code:: json

        {
          "server_group": {
            "value": {
              "server_group_node_ips": [
                "1.2.3.4",
                "1.2.3.5",
                "1.2.3.6"
              ]
            }
          }
        }

    The IP list ``server_group_node_ips`` would be used like so:

    .. code:: sh

        pyinfra @terraform/server_group.value.server_group_node_ips ...

    You can also specify dictionaries to include extra data with hosts:

    .. code:: json

        {
          "server_group": {
            "value": {
              "server_group_node_ips": [
                {
                    "ssh_hostname": "1.2.3.4",
                    "ssh_user": "ssh-user"
                },
                {
                    "ssh_hostname": "1.2.3.5",
                    "ssh_user": "ssh-user"
                }
              ]
            }
          }
        }

    """

    @override
    @staticmethod
    def make_names_data(name=None):
        show_warning()

        if not name:
            # This is the default which allows one to create a Terraform output
            # "pyinfra" and directly call: pyinfra @terraform ...
            name = "pyinfra_inventory.value"

        with progress_spinner({"fetch terraform output"}):
            tf_output_raw = local.shell("terraform output -json")

        assert isinstance(tf_output_raw, str)
        tf_output = json.loads(tf_output_raw)
        tf_output = _flatten_dict(tf_output)

        tf_output_value = tf_output.get(name)
        if tf_output_value is None:
            keys = "\n".join(f"   - {k}" for k in tf_output.keys())
            raise InventoryError(f"No Terraform output with key: `{name}`, valid keys:\n{keys}")

        if not isinstance(tf_output_value, (list, dict)):
            raise InventoryError(
                "Invalid Terraform output type, should be `list`, got "
                f"`{type(tf_output_value).__name__}`",
            )

        if isinstance(tf_output_value, list):
            tf_output_value = {
                "all": tf_output_value,
            }

        for group_name, hosts in tf_output_value.items():
            if not isinstance(hosts, list):
                raise InventoryError(
                    "Invalid Terraform map value type, all values should be `list`, got "
                    f"`{type(hosts).__name__}`",
                )
            for host in hosts:
                if isinstance(host, dict):
                    name = host.pop("name", host.get("ssh_hostname"))
                    if name is None:
                        raise InventoryError(
                            "Invalid Terraform list item, missing `name` or `ssh_hostname` keys",
                        )
                    yield f"@terraform/{name}", host, ["@terraform", group_name]
                elif isinstance(host, str):
                    data = {"ssh_hostname": host}
                    yield f"@terraform/{host}", data, ["@terraform", group_name]
                else:
                    raise InventoryError(
                        "Invalid Terraform list item, should be `dict` or `str` got "
                        f"`{type(host).__name__}`",
                    )



#!/usr/bin/env python

from glob import glob
from os import environ, path

DOCS_LANGUAGE = "en"
DOCS_VERSION = environ["DOCS_VERSION"]
TEMPLATE = """
<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8" />
    <meta name="robots" content="noindex">
    <title>Redirecting to: {0}</title>
    <link rel="canonical" href="{0}"/>
    <meta http-equiv="refresh" content="0; url={0}" />
</head>
</html>
""".lstrip()


def generate_redirect_pages():
    this_dir = path.dirname(path.realpath(__file__))
    docs_dir = path.abspath(path.join(this_dir, "..", "docs"))
    public_dir = path.join(docs_dir, "public")

    target_docs = path.join(public_dir, DOCS_LANGUAGE, DOCS_VERSION)
    files = glob(path.join(target_docs, "*.html"))

    for file in files:
        file = path.basename(file)
        filename = path.join(public_dir, "page", file)

        with open(filename, "w", encoding="utf-8") as f:
            f.write(TEMPLATE.format(f"/{DOCS_LANGUAGE}/{DOCS_VERSION}/{file}"))

        print(f"Written: {filename}")


if __name__ == "__main__":
    print("### Building operations docs")
    generate_redirect_pages()



"""
Manage runit services.
"""

from __future__ import annotations

from typing import Optional

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.files import File
from pyinfra.facts.runit import RunitManaged, RunitStatus

from .files import file, link
from .util.service import handle_service_control


@operation()
def service(
    service: str,
    running: bool = True,
    restarted: bool = False,
    reloaded: bool = False,
    command: Optional[str] = None,
    enabled: Optional[bool] = None,
    managed: bool = True,
    svdir: str = "/var/service",
    sourcedir: str = "/etc/sv",
):
    """
    Manage the state of runit services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command to pass like: ``sv <command> <service>``
    + enabled: whether this service should be enabled/disabled on boot
    + managed: whether runit should manage this service

      For services to be controlled, they first need to be managed by runit by
      adding a symlink to the service in ``SVDIR``.
      By setting ``managed=False`` the symlink will be removed.
      Other options won't have any effect after that.
      Although the ``<service>/down`` file can still be controlled with the
      ``enabled`` option.

    + svdir: alternative ``SVDIR``

      An alternative ``SVDIR`` can be specified. This can be used for user services.

    + sourcedir: where to search for available services

      An alternative directory for available services can be specified.
      Example: ``sourcedir=/etc/sv.local`` for services managed by the administrator.
    """

    was_managed = service in host.get_fact(RunitManaged, service=service, svdir=svdir)
    was_auto = not host.get_fact(File, path="{0}/{1}/down".format(sourcedir, service))

    # Disable autostart for previously unmanaged services.
    #
    # Where ``running=False`` is requested, this prevents one case of briefly
    # starting and stopping the service.
    if not was_managed and managed and was_auto:
        yield from auto._inner(
            service=service,
            auto=False,
            sourcedir=sourcedir,
        )

    yield from manage._inner(
        service=service,
        managed=managed,
        svdir=svdir,
        sourcedir=sourcedir,
    )

    # Service wasn't managed before, so wait for ``runsv`` to start.
    # ``runsvdir`` will check at least every 5 seconds for new services.
    # Wait for at most 10 seconds for the service to be managed, otherwise fail.
    if not was_managed and managed:
        yield from wait_runsv._inner(
            service=service,
            svdir=svdir,
        )

    if isinstance(enabled, bool):
        yield from auto._inner(
            service=service,
            auto=enabled,
            sourcedir=sourcedir,
        )
    else:
        # restore previous state of ``<service>/down``
        yield from auto._inner(
            service=service,
            auto=was_auto,
            sourcedir=sourcedir,
        )

    # Services need to be managed by ``runit`` for the other options to make sense.
    if not managed:
        return

    yield from handle_service_control(
        host,
        service,
        host.get_fact(RunitStatus, service=service, svdir=svdir),
        "SVDIR={0} sv {{1}} {{0}}".format(svdir),
        running,
        restarted,
        reloaded,
        command,
    )


@operation()
def manage(
    service: str,
    managed: bool = True,
    svdir: str = "/var/service",
    sourcedir: str = "/etc/sv",
):
    """
    Manage runit svdir links.

    + service: name of the service to manage
    + managed: whether the link should exist
    + svdir: alternative ``SVDIR``
    + sourcedir: where to search for available services
    """

    yield from link._inner(
        path="{0}/{1}".format(svdir, service),
        target="{0}/{1}".format(sourcedir, service),
        present=managed,
        create_remote_dir=False,
    )


@operation(is_idempotent=False)
def wait_runsv(
    service: str,
    svdir: str = "/var/service",
    timeout: int = 10,
):
    """
    Wait for runsv for ``service`` to be available.

    + service: name of the service to manage
    + svdir: alternative ``SVDIR``
    + timeout: time in seconds to wait
    """

    yield (
        "export SVDIR={0}\n"
        "for i in $(seq {1}); do\n"
        "    sv status {2} > /dev/null && exit 0\n"
        "    sleep 1;\n"
        "done\n"
        "exit 1"
    ).format(svdir, timeout, service)


@operation()
def auto(
    service: str,
    auto: bool = True,
    sourcedir: str = "/etc/sv",
):
    """
    Start service automatically by managing the ``service/down`` file.

    + service: name of the service to manage
    + auto: whether the service should start automatically
    + sourcedir: where to search for available services
    """

    yield from file._inner(
        path="{0}/{1}/down".format(sourcedir, service),
        present=not auto,
        create_remote_dir=False,
    )



from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from pyinfra.api.operation import OperationMeta


def any_changed(*args: "OperationMeta") -> Callable[[], bool]:
    return lambda: any((meta.did_change() for meta in args))


def all_changed(*args: "OperationMeta") -> Callable[[], bool]:
    return lambda: all((meta.did_change() for meta in args))



# Import `pyinfra_cli` to trigger gevent monkey patching
import logging

import gevent.hub

import pyinfra_cli  # noqa: F401
from pyinfra import logger

logging.basicConfig(level=logging.DEBUG)
logger.setLevel(logging.DEBUG)

# Don't print out exceptions inside greenlets (because here we expect them!)
gevent.hub.Hub.NOT_ERROR = (Exception,)



from datetime import date
from os import environ, mkdir, path
from shutil import rmtree

import guzzle_sphinx_theme

from pyinfra import __version__, local
from pyinfra.api import metadata

copyright = "Nick Barrett {0} — pyinfra v{1}".format(
    date.today().year,
    __version__,
)

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "guzzle_sphinx_theme",
    "myst_parser",
]
autosectionlabel_prefix_document = True

source_suffix = [".rst", ".md"]

master_doc = "index"
project = "pyinfra"
author = "Fizzadar"

version = environ.get("DOCS_VERSION", "latest")
language = "en"

pygments_style = "monokai"

# Theme style override
html_short_title = "Home"
html_theme = "guzzle_sphinx_theme"
html_theme_path = guzzle_sphinx_theme.html_theme_path()
html_static_path = ["static"]
html_theme_options = {
    "docsearch_api_key": "25a25b5f5310f306641f9ce07dcb06eb",
    "docsearch_app_id": "XXGX6EX4KA",
    "docsearch_index_name": "pyinfra",
    "plausible_domain": "docs.pyinfra.com",
    "plausible_stats_domain": "stats.oxygem.com",
    "doc_versions": ["3.x", "2.x", "1.x", "0.x", "latest"],
    "primary_doc_version": "3.x",
}

myst_heading_anchors = 3

templates_path = ["templates"]

html_favicon = "static/logo_small.png"

exclude_patterns = [
    "_deploy_globals.rst",
]


def rstjinja(app, docname, source):
    """
    Render certain pages as a jinja templates.
    """
    # this should only be run when building html
    if app.builder.format != "html":
        return
    # We currently only render docs/operations.rst
    # and docs/facts.rst as jinja2 templates
    if docname in ["operations", "facts"]:
        src = source[0]
        rendered = app.builder.templates.render_string(src, app.config.html_context)
        source[0] = rendered


def setup(app):
    app.connect("source-read", rstjinja)
    this_dir = path.dirname(path.realpath(__file__))
    scripts_dir = path.abspath(path.join(this_dir, "..", "scripts"))
    metadata_file = path.abspath(path.join(this_dir, "..", "pyinfra-metadata.toml"))
    if not path.exists(metadata_file):
        raise ValueError("No pyinfra-metadata.toml in project root")
    with open(metadata_file, "r") as file:
        metadata_text = file.read()
    plugins = metadata.parse_plugins(metadata_text)

    operation_plugins = sorted([p for p in plugins if p.type == "operation"], key=lambda p: p.name)
    fact_plugins = sorted([p for p in plugins if p.type == "fact"], key=lambda p: p.name)
    html_context = {
        "operation_plugins": operation_plugins,
        "fact_plugins": fact_plugins,
        "tags": metadata.ALLOWED_TAGS,
        "docs_language": language,
        "docs_version": version,
    }
    app.config.html_context = html_context

    for auto_docs_name in ("operations", "facts", "apidoc", "connectors"):
        auto_docs_path = path.join(this_dir, auto_docs_name)
        if path.exists(auto_docs_path):
            rmtree(auto_docs_path)
        mkdir(auto_docs_path)

    local.shell(
        (
            "python {0}/generate_api_docs.py".format(scripts_dir),
            "python {0}/generate_arguments_doc.py".format(scripts_dir),
            "python {0}/generate_connectors_docs.py".format(scripts_dir),
            "python {0}/generate_facts_docs.py".format(scripts_dir),
            "python {0}/generate_operations_docs.py".format(scripts_dir),
        ),
        print_input=True,
    )



from os import path

from utils import call_file_op

from pyinfra import host, local
from pyinfra.api import deploy
from pyinfra.operations import files, server


@deploy("My nested deploy")
def my_nested_deploy():
    server.shell(
        name="First nested deploy operation",
        commands="echo first nested_deploy_op",
    )


@deploy("My deploy")
def my_deploy():
    server.shell(
        name="First deploy operation",
        commands="echo first_deploy_op",
    )

    my_nested_deploy()

    server.shell(
        name="Second deploy operation",
        commands="echo second_deploy_op",
    )


server.shell(
    name="First main operation",
    commands="echo first_main_op",
)

# Create some conditional branches
if host.name == "somehost":
    server.shell(
        name="Second main operation",
        commands="echo second_main_op",
    )
elif host.name == "anotherhost":
    local.include(path.join("tasks", "a_task.py"))

# Include the whole file again, but for all hosts
local.include(path.join("tasks", "a_task.py"))

# Include a deploy file, with custom specified data
local.include(path.join("tasks", "b_task.py"), data={"keyword": "Important", "id": 1})

# Execute the @deploy function
my_deploy()

# Do a loop which will generate duplicate op hashes
for i in range(2):
    server.shell(
        name="Loop-{0} main operation".format(i),
        commands="echo loop_{0}_main_operation".format(i),
    )

call_file_op()

# Ensure complex nested loops don't generate cycles
for item in host.loop([1, 2]):
    server.shell(
        name=f"Order loop {item}",
        commands=f"echo loop_{item}",
    )
    if host.name == "anotherhost" or (item == 2 and host.name == "somehost"):
        for inner_item in host.loop([1, 2]):
            server.shell(
                name=f"Nested order loop {item}/{inner_item}",
                commands=f"echo loop_{item}/{inner_item}",
            )

if host.name == "somehost":
    files.template(
        name="Final limited operation",
        src="templates/a_template.j2",
        dest="/a_template",
        is_template=True,
    )

with host.when(lambda: host.name != "somehost"):
    server.shell(
        name="Second final limited operation",
        commands="echo final_limited_op",
    )



"""
Manage apt packages and repositories.
"""

from __future__ import annotations

from datetime import datetime, timedelta, timezone
from urllib.parse import urlparse

from pyinfra import host
from pyinfra.api import OperationError, operation
from pyinfra.facts.apt import (
    AptKeys,
    AptSources,
    SimulateOperationWillChange,
    noninteractive_apt,
    parse_apt_repo,
)
from pyinfra.facts.deb import DebPackage, DebPackages
from pyinfra.facts.files import File
from pyinfra.facts.gpg import GpgKey
from pyinfra.facts.server import Date

from . import files
from .util.packaging import ensure_packages

APT_UPDATE_FILENAME = "/var/lib/apt/periodic/update-success-stamp"


def _simulate_then_perform(command: str):
    changes = host.get_fact(SimulateOperationWillChange, command)

    if not changes:
        # Simulating apt-get command failed, so the actual
        # operation will probably fail too:
        yield noninteractive_apt(command)
    elif (
        changes["upgraded"] == 0
        and changes["newly_installed"] == 0
        and changes["removed"] == 0
        and changes["not_upgraded"] == 0
    ):
        host.noop(f"{command} skipped, no changes would be performed")
    else:
        yield noninteractive_apt(command)


@operation()
def key(src: str | None = None, keyserver: str | None = None, keyid: str | list[str] | None = None):
    """
    Add apt gpg keys with ``apt-key``.

    + src: filename or URL
    + keyserver: URL of keyserver to fetch key from
    + keyid: key ID or list of key IDs when using keyserver

    keyserver/id:
        These must be provided together.

    .. warning::
        ``apt-key`` is deprecated in Debian, it is recommended NOT to use this
        operation and instead follow the instructions here:

        https://wiki.debian.org/DebianRepository/UseThirdParty

    **Examples:**

    .. code:: python

        from pyinfra.operations import apt
        # Note: If using URL, wget is assumed to be installed.
        apt.key(
            name="Add the Docker apt gpg key",
            src="https://download.docker.com/linux/ubuntu/gpg",
        )

        apt.key(
            name="Install VirtualBox key",
            src="https://www.virtualbox.org/download/oracle_vbox_2016.asc",
        )
    """

    existing_keys = host.get_fact(AptKeys)

    if src:
        key_data = host.get_fact(GpgKey, src=src)
        if key_data:
            keyid = list(key_data.keys())

        if not keyid or not all(kid in existing_keys for kid in keyid):
            # If URL, wget the key to stdout and pipe into apt-key, because the "adv"
            # apt-key passes to gpg which doesn't always support https!
            if urlparse(src).scheme:
                yield "(wget -O - {0} || curl -sSLf {0}) | apt-key add -".format(src)
            else:
                yield "apt-key add {0}".format(src)
        else:
            host.noop("All keys from {0} are already available in the apt keychain".format(src))

    if keyserver:
        if not keyid:
            raise OperationError("`keyid` must be provided with `keyserver`")

        if isinstance(keyid, str):
            keyid = [keyid]

        needed_keys = sorted(set(keyid) - set(existing_keys.keys()))
        if needed_keys:
            yield "apt-key adv --keyserver {0} --recv-keys {1}".format(
                keyserver,
                " ".join(needed_keys),
            )
        else:
            host.noop(
                "Keys {0} are already available in the apt keychain".format(
                    ", ".join(keyid),
                ),
            )


@operation()
def repo(src: str, present=True, filename: str | None = None):
    """
    Add/remove apt repositories.

    + src: apt source string eg ``deb http://X hardy main``
    + present: whether the repo should exist on the system
    + filename: optional filename to use ``/etc/apt/sources.list.d/<filename>.list``. By
      default uses ``/etc/apt/sources.list``.

    **Example:**

    .. code:: python

        apt.repo(
            name="Install VirtualBox repo",
            src="deb https://download.virtualbox.org/virtualbox/debian bionic contrib",
        )
    """

    # Get the target .list file to manage
    if filename:
        filename = "/etc/apt/sources.list.d/{0}.list".format(filename)
    else:
        filename = "/etc/apt/sources.list"

    # Work out if the repo exists already
    apt_sources = host.get_fact(AptSources)

    is_present = False
    repo = parse_apt_repo(src)
    if repo and repo in apt_sources:
        is_present = True

    # Doesn't exist and we want it
    if not is_present and present:
        yield from files.line._inner(
            path=filename,
            line=src,
            escape_regex_characters=True,
        )

    # Exists and we don't want it
    elif is_present and not present:
        yield from files.line._inner(
            path=filename,
            line=src,
            present=False,
            escape_regex_characters=True,
        )
    else:
        host.noop(
            'apt repo "{0}" {1}'.format(
                src,
                "exists" if present else "does not exist",
            ),
        )


@operation(is_idempotent=False)
def ppa(src: str, present=True):
    """
    Add/remove Ubuntu ppa repositories.

    + src: the PPA name (full ppa:user/repo format)
    + present: whether it should exist

    Note:
        requires ``apt-add-repository`` on the remote host

    **Example:**

    .. code:: python

        # Note: Assumes software-properties-common is installed.
        apt.ppa(
            name="Add the Bitcoin ppa",
            src="ppa:bitcoin/bitcoin",
        )

    """

    if present:
        yield 'apt-add-repository -y "{0}"'.format(src)

    if not present:
        yield 'apt-add-repository -y --remove "{0}"'.format(src)


@operation()
def deb(src: str, present=True, force=False):
    """
    Add/remove ``.deb`` file packages.

    + src: filename or URL of the ``.deb`` file
    + present: whether or not the package should exist on the system
    + force: whether to force the package install by passing `--force-yes` to apt

    Note:
        When installing, ``apt-get install -f`` will be run to install any unmet
        dependencies.

    URL sources with ``present=False``:
        If the ``.deb`` file isn't downloaded, pyinfra can't remove any existing
        package as the file won't exist until mid-deploy.

    **Example:**

    .. code:: python

        # Note: Assumes wget is installed.
        apt.deb(
            name="Install Chrome via deb",
            src="https://dl.google.com/linux/direct/google-chrome-stable_current_amd64.deb",
        )
    """

    original_src = src

    # If source is a url
    if urlparse(src).scheme:
        # Generate a temp filename
        temp_filename = host.get_temp_filename(src)

        # Ensure it's downloaded
        yield from files.download._inner(src=src, dest=temp_filename)

        # Override the source with the downloaded file
        src = temp_filename

    # Check for file .deb information (if file is present)
    info = host.get_fact(DebPackage, package=src)
    current_packages = host.get_fact(DebPackages)

    exists = False

    # We have deb info! Check against installed packages
    if info and info.get("version") in current_packages.get(info.get("name"), {}):
        exists = True

    # Package does not exist and we want?
    if present:
        if not exists:
            # Install .deb file - ignoring failure (on unmet dependencies)
            yield "dpkg --force-confdef --force-confold -i {0} 2> /dev/null || true".format(src)
            # Attempt to install any missing dependencies
            yield "{0} -f".format(noninteractive_apt("install", force=force))
            # Now reinstall, and critically configure, the package - if there are still
            # missing deps, now we error
            yield "dpkg --force-confdef --force-confold -i {0}".format(src)
        else:
            host.noop("deb {0} is installed".format(original_src))

    # Package exists but we don't want?
    if not present:
        if exists:
            yield "{0} {1}".format(
                noninteractive_apt("remove", force=force),
                info["name"],
            )
        else:
            host.noop("deb {0} is not installed".format(original_src))


@operation(
    is_idempotent=False,
    idempotent_notice=(
        "This operation will always execute commands "
        "unless the ``cache_time`` argument is provided."
    ),
)
def update(cache_time: int | None = None):
    """
    Updates apt repositories.

    + cache_time: cache updates for this many seconds

    **Example:**

    .. code:: python

        apt.update(
            name="Update apt repositories",
            cache_time=3600,
        )
    """

    # If cache_time check when apt was last updated, prevent updates if within time
    if cache_time:
        # Ubuntu provides this handy file
        cache_info = host.get_fact(File, path=APT_UPDATE_FILENAME)

        if cache_info and cache_info["mtime"]:
            # The fact Date contains the date of the server in its timezone.
            # cache_info["mtime"] ignores the timezone and consider the timestamp as UTC.
            # So let's do the same here for the server current Date : ignore the
            # timezone and consider it as UTC to have correct comparison with
            # cache_info["mtime].
            host_utc_current_time = datetime.fromtimestamp(
                host.get_fact(Date).timestamp(), timezone.utc
            ).replace(tzinfo=None)
            host_cache_time = host_utc_current_time - timedelta(seconds=cache_time)
            if cache_info["mtime"] > host_cache_time:
                host.noop("apt is already up to date")
                return

    yield "apt-get update"

    # Some apt systems (Debian) have the /var/lib/apt/periodic directory, but
    # don't bother touching anything in there - so pyinfra does it, enabling
    # cache_time to work.
    if cache_time:
        yield "touch {0}".format(APT_UPDATE_FILENAME)


_update = update  # noqa: E305


@operation()
def upgrade(auto_remove: bool = False):
    """
    Upgrades all apt packages.

    + auto_remove: removes transitive dependencies that are no longer needed.

    **Example:**

    .. code:: python

        # Upgrade all packages
        apt.upgrade(
            name="Upgrade apt packages",
        )

        # Upgrade all packages and remove unneeded transitive dependencies
        apt.upgrade(
            name="Upgrade apt packages and remove unneeded dependencies",
            auto_remove=True
        )
    """

    command = ["upgrade"]

    if auto_remove:
        command.append("--autoremove")

    yield from _simulate_then_perform(" ".join(command))


_upgrade = upgrade  # noqa: E305 (for use below where update is a kwarg)


@operation()
def dist_upgrade(auto_remove: bool = False):
    """
    Updates all apt packages, employing dist-upgrade.

    + auto_remove: removes transitive dependencies that are no longer needed.

    **Example:**

    .. code:: python

        apt.dist_upgrade(
            name="Upgrade apt packages using dist-upgrade",
        )
    """

    command = ["dist-upgrade"]

    if auto_remove:
        command.append("--autoremove")

    yield from _simulate_then_perform(" ".join(command))


@operation()
def packages(
    packages: str | list[str] | None = None,
    present=True,
    latest=False,
    update=False,
    cache_time: int | None = None,
    upgrade=False,
    force=False,
    no_recommends=False,
    allow_downgrades=False,
    extra_install_args: str | None = None,
    extra_uninstall_args: str | None = None,
):
    """
    Install/remove/update packages & update apt.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``apt update`` before installing packages
    + cache_time: when used with ``update``, cache for this many seconds
    + upgrade: run ``apt upgrade`` before installing packages
    + force: whether to force package installs by passing `--force-yes` to apt
    + no_recommends: don't install recommended packages
    + allow_downgrades: allow downgrading packages with version (--allow-downgrades)
    + extra_install_args: additional arguments to the apt install command
    + extra_uninstall_args: additional arguments to the apt uninstall command

    Versions:
        Package versions can be pinned like apt: ``<pkg>=<version>``

    Cache time:
        When ``cache_time`` is set the ``/var/lib/apt/periodic/update-success-stamp`` file
        is touched upon successful update. Some distros already do this (Ubuntu), but others
        simply leave the periodic directory empty (Debian).

    **Examples:**

    .. code:: python

        # Update package list and install packages
        apt.packages(
            name="Install Asterisk and Vim",
            packages=["asterisk", "vim"],
            update=True,
        )

        # Install the latest versions of packages (always check)
        apt.packages(
            name="Install latest Vim",
            packages=["vim"],
            latest=True,
        )

        # Note: host.get_fact(OsVersion) is the same as `uname -r` (ex: '4.15.0-72-generic')
        apt.packages(
            name="Install kernel headers",
            packages=[f"linux-headers-{host.get_fact(OsVersion)}"],
            update=True,
        )
    """

    if update:
        yield from _update._inner(cache_time=cache_time)

    if upgrade:
        yield from _upgrade._inner()

    install_command_args = ["install"]
    if no_recommends is True:
        install_command_args.append("--no-install-recommends")
    if allow_downgrades:
        install_command_args.append("--allow-downgrades")

    upgrade_command = " ".join(install_command_args)

    if extra_install_args:
        install_command_args.append(extra_install_args)

    install_command = " ".join(install_command_args)

    uninstall_command_args = ["remove"]
    if extra_uninstall_args:
        uninstall_command_args.append(extra_uninstall_args)

    uninstall_command = " ".join(uninstall_command_args)

    # Compare/ensure packages are present/not
    yield from ensure_packages(
        host,
        packages,
        host.get_fact(DebPackages),
        present,
        install_command=noninteractive_apt(install_command, force=force),
        uninstall_command=noninteractive_apt(uninstall_command, force=force),
        upgrade_command=noninteractive_apt(upgrade_command, force=force),
        version_join="=",
        latest=latest,
    )



#!/usr/bin/env python

from os import path


def cleanup_words():
    with open(path.join("tests", "words.txt"), "r", encoding="utf-8") as f:
        data = f.read()

    lines = data.splitlines()
    lines = [line for line in lines if not line.startswith("#")]

    lines = sorted(set(lines))

    lines.insert(0, "# it is automatically cleaned/sorted by scripts/cleanup_words.py")
    lines.insert(0, "# This is a list of additional words for flake8-spellcheck")

    with open(path.join("tests", "words.txt"), "w", encoding="utf-8") as f:
        f.write("\n".join(lines))
        f.write("\n")


if __name__ == "__main__":
    cleanup_words()



#!/usr/bin/env python
"""
Lint script to verify that arguments in arguments.py and arguments_typed.py are in sync.

Due to Python's typing limitations (ParamSpec args cannot be concatenated), we maintain
two separate definitions of global arguments:
- arguments.py: TypedDicts (ConnectorArguments, MetaArguments, ExecutionArguments)
- arguments_typed.py: PyinfraOperation.__call__ method signature

This script ensures they stay synchronized.
"""

import ast
import sys
from os import path
from typing import NamedTuple


class ArgumentInfo(NamedTuple):
    name: str
    type_annotation: str
    has_default: bool


def get_typeddict_keys(tree: ast.Module, class_names: list[str]) -> dict[str, str]:
    """Extract keys and their type annotations from TypedDict classes."""
    keys: dict[str, str] = {}

    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef) and node.name in class_names:
            for item in node.body:
                if isinstance(item, ast.AnnAssign) and isinstance(item.target, ast.Name):
                    key = item.target.id
                    type_str = ast.unparse(item.annotation)
                    keys[key] = type_str

    return keys


def get_argument_meta_keys(tree: ast.Module) -> set[str]:
    """Extract keys from all *_argument_meta dictionaries."""
    keys: set[str] = set()

    for node in ast.walk(tree):
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id.endswith("_argument_meta"):
                    if isinstance(node.value, ast.Dict):
                        for key in node.value.keys:
                            if isinstance(key, ast.Constant) and isinstance(key.value, str):
                                keys.add(key.value)

    return keys


def get_call_parameters(tree: ast.Module, class_name: str, method_name: str) -> list[ArgumentInfo]:
    """Extract parameters from a class method."""
    params: list[ArgumentInfo] = []

    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef) and node.name == class_name:
            for item in node.body:
                if isinstance(item, ast.FunctionDef) and item.name == method_name:
                    args = item.args

                    # Count defaults - they align with the end of the args list
                    num_defaults = len(args.defaults)
                    num_args = len(args.args)

                    for i, arg in enumerate(args.args):
                        if arg.arg == "self":
                            continue

                        type_str = ast.unparse(arg.annotation) if arg.annotation else ""

                        # Check if this arg has a default
                        default_index = i - (num_args - num_defaults)
                        has_default = default_index >= 0

                        params.append(
                            ArgumentInfo(
                                name=arg.arg,
                                type_annotation=type_str,
                                has_default=has_default,
                            )
                        )

                    # Also check kwonly args
                    for i, arg in enumerate(args.kwonlyargs):
                        type_str = ast.unparse(arg.annotation) if arg.annotation else ""
                        has_default = args.kw_defaults[i] is not None
                        params.append(
                            ArgumentInfo(
                                name=arg.arg,
                                type_annotation=type_str,
                                has_default=has_default,
                            )
                        )

    return params


def normalize_type(type_str: str) -> str:
    """Normalize type string for comparison (handle Optional, Union, etc.)."""
    # Remove whitespace
    type_str = type_str.replace(" ", "")

    # Sort Union members for consistent comparison
    if type_str.startswith("Union[") or type_str.startswith("Optional["):
        # This is a simplified normalization - just for basic comparison
        pass

    return type_str


def main() -> int:
    this_dir = path.dirname(path.realpath(__file__))
    repo_root = path.abspath(path.join(this_dir, ".."))

    arguments_path = path.join(repo_root, "src", "pyinfra", "api", "arguments.py")
    arguments_typed_path = path.join(repo_root, "src", "pyinfra", "api", "arguments_typed.py")

    # Parse both files
    with open(arguments_path, "r", encoding="utf-8") as f:
        arguments_tree = ast.parse(f.read())

    with open(arguments_typed_path, "r", encoding="utf-8") as f:
        arguments_typed_tree = ast.parse(f.read())

    # Extract TypedDict keys from arguments.py
    typeddict_classes = ["ConnectorArguments", "MetaArguments", "ExecutionArguments"]
    typeddict_keys = get_typeddict_keys(arguments_tree, typeddict_classes)

    # Extract argument meta keys (the actual source of truth for what arguments exist)
    meta_keys = get_argument_meta_keys(arguments_tree)

    # Extract PyinfraOperation.__call__ parameters from arguments_typed.py
    call_params = get_call_parameters(arguments_typed_tree, "PyinfraOperation", "__call__")
    call_param_names = {p.name for p in call_params}
    call_param_types = {p.name: p.type_annotation for p in call_params}

    errors: list[str] = []
    warnings: list[str] = []

    # Check that all TypedDict keys are in PyinfraOperation.__call__
    for key, type_str in typeddict_keys.items():
        if key not in call_param_names:
            errors.append(
                f"Argument '{key}' is in arguments.py TypedDicts but missing from "
                f"PyinfraOperation.__call__ in arguments_typed.py"
            )
        elif key in call_param_types:
            typed_type = normalize_type(call_param_types[key])
            expected_type = normalize_type(type_str)

            # TypedDict uses non-Optional types, but PyinfraOperation uses Optional
            # So we do a loose check - the base type should match
            # This is a simplified check that may need refinement
            if expected_type not in typed_type and typed_type not in expected_type:
                # Check if it's just an Optional wrapper difference
                if f"Optional[{expected_type}]" != typed_type and expected_type != typed_type:
                    warnings.append(
                        f"Type mismatch for '{key}': "
                        f"arguments.py has '{type_str}', "
                        f"arguments_typed.py has '{call_param_types[key]}'"
                    )

    # Check that all PyinfraOperation.__call__ params (except special ones) are in TypedDicts
    # Skip *args and **kwargs which are the P.args/P.kwargs for operation-specific args
    special_params = {"args", "kwargs"}
    for param in call_params:
        if param.name in special_params:
            continue
        if param.name not in typeddict_keys:
            errors.append(
                f"Parameter '{param.name}' is in PyinfraOperation.__call__ but missing from "
                f"TypedDicts in arguments.py"
            )

    # Check that all meta keys are represented
    all_typeddict_keys = set(typeddict_keys.keys())
    for key in meta_keys:
        if key not in all_typeddict_keys:
            errors.append(f"Argument '{key}' is in argument_meta dicts but missing from TypedDicts")

    # Report results
    if warnings:
        print("Warnings:")
        for warning in warnings:
            print(f"  ⚠️  {warning}")
        print()

    if errors:
        print("Errors:")
        for error in errors:
            print(f"  ❌ {error}")
        print()
        print(f"Found {len(errors)} error(s) and {len(warnings)} warning(s)")
        return 1

    print("✅ arguments.py and arguments_typed.py are in sync!")
    if warnings:
        print(f"   ({len(warnings)} warning(s) - type annotations may differ slightly)")
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

from os import path, remove

from pyinfra import local


def generate_api_docs():
    this_dir = path.dirname(path.realpath(__file__))
    docs_dir = path.abspath(path.join(this_dir, "..", "docs"))
    pyinfra_dir = path.abspath(path.join(this_dir, "..", "src", "pyinfra"))

    api_doc_command = (
        "uv run sphinx-apidoc -e -M -f -o {0}/apidoc/ {1} {1}/facts {1}/operations {1}/connectors"
    ).format(
        docs_dir,
        pyinfra_dir,
    )

    local.shell(
        (api_doc_command,),
        print_input=True,
    )

    for filename in ("modules.rst", "pyinfra.rst", "pyinfra.api.rst"):
        remove("{0}/apidoc/{1}".format(docs_dir, filename))


if __name__ == "__main__":
    print("### Generating API docs")
    generate_api_docs()



from __future__ import annotations

import difflib
import re
from datetime import datetime, timezone
from enum import Enum
from typing import Callable, Generator

import click

from pyinfra.api import QuoteString, StringCommand


class MetadataTimeField(Enum):
    ATIME = "atime"
    MTIME = "mtime"


def unix_path_join(*parts) -> str:
    part_list = list(parts)
    part_list[0:-1] = [part.rstrip("/") for part in part_list[0:-1]]
    return "/".join(part_list)


def ensure_mode_int(mode: str | int | None) -> int | str | None:
    # Already an int (/None)?
    if isinstance(mode, int) or mode is None:
        return mode

    try:
        # Try making an int ('700' -> 700)
        return int(mode)

    except (TypeError, ValueError):
        pass

    # Return as-is (ie +x which we don't need to normalise, it always gets run)
    return mode


def get_timestamp() -> str:
    return datetime.now().strftime("%y%m%d%H%M")


_sed_ignore_case = re.compile("[iI]")


def _sed_delete_builder(line: str, replace: str, flags: str, interpolate_variables: bool) -> str:
    return (
        '"/{0}/{1}d"'
        if interpolate_variables
        # fmt: skip
        else "'/{0}/{1}d'"
    ).format(line, "I" if _sed_ignore_case.search(flags) else "")


def sed_delete(
    filename: str,
    line: str,
    replace: str,
    flags: list[str] | None = None,
    backup=False,
    interpolate_variables=False,
) -> StringCommand:
    return _sed_command(**locals(), sed_script_builder=_sed_delete_builder)


def _sed_replace_builder(line: str, replace: str, flags: str, interpolate_variables: bool) -> str:
    return (
        '"s/{0}/{1}/{2}"'
        if interpolate_variables
        # fmt: skip
        else "'s/{0}/{1}/{2}'"
    ).format(line, replace, flags)


def sed_replace(
    filename: str,
    line: str,
    replace: str,
    flags: list[str] | None = None,
    backup=False,
    interpolate_variables=False,
) -> StringCommand:
    return _sed_command(**locals(), sed_script_builder=_sed_replace_builder)


def _sed_command(
    filename: str,
    line: str,
    replace: str,
    flags: list[str] | None = None,
    backup=False,
    interpolate_variables=False,
    # Python requires a default value here, so use _sed_replace_builder for
    # backwards compatibility.
    sed_script_builder: Callable[[str, str, str, bool], str] = _sed_replace_builder,
) -> StringCommand:
    flags_str = "".join(flags) if flags else ""

    line = line.replace("/", r"\/")
    replace = str(replace)
    replace = replace.replace("/", r"\/")
    replace = replace.replace("&", r"\&")
    backup_extension = get_timestamp()

    if interpolate_variables:
        line = line.replace('"', '\\"')
        replace = replace.replace('"', '\\"')
    else:
        # Single quotes cannot contain other single quotes, even when escaped , so turn
        # each ' into '"'"' (end string, double quote the single quote, (re)start string)
        line = line.replace("'", "'\"'\"'")
        replace = replace.replace("'", "'\"'\"'")

    sed_script = sed_script_builder(line, replace, flags_str, interpolate_variables)

    sed_command = StringCommand(
        "sed",
        "-i.{0}".format(backup_extension),
        sed_script,
        QuoteString(filename),
    )

    if not backup:  # if we're not backing up, remove the file *if* sed succeeds
        backup_filename = "{0}.{1}".format(filename, backup_extension)
        sed_command = StringCommand(sed_command, "&&", "rm", "-f", QuoteString(backup_filename))

    return sed_command


def chmod(target: str, mode: str | int, recursive=False) -> StringCommand:
    args = ["chmod"]
    if recursive:
        args.append("-R")

    args.append("{0}".format(mode))

    return StringCommand(" ".join(args), QuoteString(target))


def chown(
    target: str,
    user: str | None = None,
    group: str | None = None,
    recursive=False,
    dereference=True,
) -> StringCommand:
    command = "chown"
    user_group = None

    if user and group:
        user_group = "{0}:{1}".format(user, group)

    elif user:
        user_group = user

    elif group:
        command = "chgrp"
        user_group = group

    args = [command]
    if recursive:
        args.append("-R")

    if not dereference:
        args.append("-h")

    return StringCommand(" ".join(args), user_group, QuoteString(target))


# like the touch command, but only supports setting one field at a time, and expects any
#   reference times to have been read from the reference file metadata and turned into
#   aware datetimes
def touch(
    target: str,
    timefield: MetadataTimeField,
    timesrc: datetime,
    dereference=True,
) -> StringCommand:
    args = ["touch"]

    if timefield is MetadataTimeField.ATIME:
        args.append("-a")
    else:
        args.append("-m")

    if not dereference:
        args.append("-h")

    # don't reinvent the wheel; use isoformat()
    timestr = timesrc.astimezone(timezone.utc).isoformat()
    # but replace the ISO format TZ offset with "Z" for BSD
    timestr = timestr.replace("+00:00", "Z")
    args.extend(["-d", timestr])

    return StringCommand(" ".join(args), QuoteString(target))


def adjust_regex(line: str, escape_regex_characters: bool) -> str:
    """
    Ensure the regex starts with '^' and ends with '$' and escape regex characters if requested
    """
    match_line = line

    if escape_regex_characters:
        match_line = re.sub(r"([\.\\\+\*\?\[\^\]\$\(\)\{\}\-])", r"\\\1", match_line)

    # Ensure we're matching a whole line, note: match may be a partial line so we
    # put any matches on either side.
    if not match_line.startswith("^"):
        match_line = "^.*{0}".format(match_line)
    if not match_line.endswith("$"):
        match_line = "{0}.*$".format(match_line)

    return match_line


def generate_color_diff(
    current_lines: list[str], desired_lines: list[str]
) -> Generator[str, None, None]:
    def _format_range_unified(start: int, stop: int) -> str:
        beginning = start + 1  # lines start numbering with one
        length = stop - start
        if length == 1:
            return "{}".format(beginning)
        if not length:
            beginning -= 1  # empty ranges begin at line just before the range
        return "{},{}".format(beginning, length)

    for group in difflib.SequenceMatcher(None, current_lines, desired_lines).get_grouped_opcodes(2):
        first, last = group[0], group[-1]
        file1_range = _format_range_unified(first[1], last[2])
        file2_range = _format_range_unified(first[3], last[4])
        yield "@@ -{} +{} @@".format(file1_range, file2_range)

        for tag, i1, i2, j1, j2 in group:
            if tag == "equal":
                for line in current_lines[i1:i2]:
                    yield "  " + line.rstrip()
                continue
            if tag in {"replace", "delete"}:
                for line in current_lines[i1:i2]:
                    yield click.style("- " + line.rstrip(), "red")
            if tag in {"replace", "insert"}:
                for line in desired_lines[j1:j2]:
                    yield click.style("+ " + line.rstrip(), "green")



from pyinfra.facts.util.units import parse_human_readable_size


def test_parse_human_readable_size():
    example_strings = [
        "1024b",
        "10.43 KB",
        "11 GB",
        "343.1 MB",
        "10.43KB",
        "11GB",
        "343.1MB",
        "10.43 kb",
        "11 gb",
        "343.1 mb",
        "10.43kb",
        "11gb",
        "343.1mb",
        "1024Kib",
        "10.43 KiB",
        "11 GiB",
        "343.1 MiB",
        "10.43KiB",
        "11GiB",
        "343.1MiB",
        "10.43 kib",
        "11 gib",
    ]
    for example_string in example_strings:
        print(example_string, parse_human_readable_size(example_string))



class PyinfraError(Exception):
    """
    Generic pyinfra exception.
    """


class ConnectError(PyinfraError):
    """
    Exception raised when connecting fails.
    """


class FactError(PyinfraError):
    """
    Exception raised during fact gathering staging if a fact is unable to
    generate output/change state.
    """


class FactTypeError(FactError, TypeError):
    """
    Exception raised when a fact is passed invalid argument types.
    """


class FactValueError(FactError, ValueError):
    """
    Exception raised when a fact is passed invalid argument values.
    """


class FactProcessError(FactError, RuntimeError):
    """
    Exception raised when the data gathered for a fact cannot be processed.
    """


class OperationError(PyinfraError):
    """
    Exception raised during fact gathering staging if an operation is unable to
    generate output/change state.
    """


class OperationTypeError(OperationError, TypeError):
    """
    Exception raised when an operation is passed invalid argument types.
    """


class OperationValueError(OperationError, ValueError):
    """
    Exception raised when an operation is passed invalid argument values.
    """


class NestedOperationError(OperationError):
    """
    Exception raised when a nested (immediately executed) operation fails.
    """


class DeployError(PyinfraError):
    """
    User exception for raising in deploys or sub deploys.
    """


class InventoryError(PyinfraError):
    """
    Exception raised for inventory related errors.
    """


class NoConnectorError(PyinfraError, ValueError):
    """
    Raised when a requested connector is missing.
    """


class NoHostError(PyinfraError, KeyError):
    """
    Raised when an inventory is missing a host.
    """


class NoGroupError(PyinfraError, KeyError):
    """
    Raised when an inventory is missing a group.
    """


class ConnectorDataTypeError(PyinfraError, TypeError):
    """
    Raised when host connector data has invalid types.
    """


class ArgumentTypeError(PyinfraError, TypeError):
    """
    Raised when global arguments are passed with invalid types.
    """



import os
from shutil import which
from tempfile import mkstemp
from typing import TYPE_CHECKING, Tuple

import click
from typing_extensions import Unpack, override

from pyinfra import logger
from pyinfra.api.command import QuoteString, StringCommand
from pyinfra.api.exceptions import InventoryError
from pyinfra.api.util import get_file_io

from .base import BaseConnector
from .util import (
    CommandOutput,
    execute_command_with_sudo_retry,
    make_unix_command_for_host,
    run_local_process,
)

if TYPE_CHECKING:
    from pyinfra.api.arguments import ConnectorArguments


class LocalConnector(BaseConnector):
    """
    The ``@local`` connector executes changes on the local machine using
    subprocesses. **This connector is only compatible with MacOS & Linux hosts**.

    Examples:

    .. code::

        # Install nginx
        pyinfra inventory.py apt.packages nginx update=true _sudo=true
    """

    handles_execution = True

    @override
    @staticmethod
    def make_names_data(name=None):
        if name is not None:
            raise InventoryError("Cannot have more than one @local")

        yield "@local", {}, ["@local"]

    @override
    def run_shell_command(
        self,
        command: StringCommand,
        print_output: bool = False,
        print_input: bool = False,
        **arguments: Unpack["ConnectorArguments"],
    ) -> Tuple[bool, CommandOutput]:
        """
        Execute a command on the local machine.

        Args:
            command (StringCommand): actual command to execute
            print_output (bool): whether to print command output
            print_input (bool): whether to print command input
            arguments: (ConnectorArguments): connector global arguments

        Returns:
            tuple: (bool, CommandOutput)
            Bool indicating success and CommandOutput with stdout/stderr lines.
        """

        arguments.pop("_get_pty", False)
        _timeout = arguments.pop("_timeout", None)
        _stdin = arguments.pop("_stdin", None)
        _success_exit_codes = arguments.pop("_success_exit_codes", None)

        def execute_command() -> Tuple[int, CommandOutput]:
            unix_command = make_unix_command_for_host(self.state, self.host, command, **arguments)
            actual_command = unix_command.get_raw_value()

            logger.debug("--> Running command on localhost: %s", unix_command)

            if print_input:
                click.echo("{0}>>> {1}".format(self.host.print_prefix, unix_command), err=True)

            return run_local_process(
                actual_command,
                stdin=_stdin,
                timeout=_timeout,
                print_output=print_output,
                print_prefix=self.host.print_prefix,
            )

        return_code, combined_output = execute_command_with_sudo_retry(
            self.host,
            arguments,
            execute_command,
        )

        if _success_exit_codes:
            status = return_code in _success_exit_codes
        else:
            status = return_code == 0

        return status, combined_output

    @override
    def put_file(
        self,
        filename_or_io,
        remote_filename,
        remote_temp_filename=None,  # ignored
        print_output: bool = False,
        print_input: bool = False,
        **arguments,
    ) -> bool:
        """
        Upload a local file or IO object by copying it to a temporary directory
        and then writing it to the upload location.

        Returns:
            bool: Indicating success or failure
        """

        _, temp_filename = mkstemp()

        try:
            # Load our file or IO object and write it to the temporary file
            with get_file_io(filename_or_io) as file_io:
                with open(temp_filename, "wb") as temp_f:
                    data = file_io.read()

                    if isinstance(data, str):
                        data = data.encode()

                    temp_f.write(data)

            # Copy the file using `cp` such that we support sudo/su
            status, output = self.run_shell_command(
                StringCommand("cp", temp_filename, QuoteString(remote_filename)),
                print_output=print_output,
                print_input=print_input,
                **arguments,
            )

            if not status:
                raise IOError(output.stderr)
        finally:
            os.remove(temp_filename)

        if print_output:
            click.echo(
                "{0}file copied: {1}".format(self.host.print_prefix, remote_filename),
                err=True,
            )

        return status

    @override
    def get_file(
        self,
        remote_filename,
        filename_or_io,
        remote_temp_filename=None,  # ignored
        print_output: bool = False,
        print_input: bool = False,
        **arguments,
    ) -> bool:
        """
        Download a local file by copying it to a temporary location and then writing
        it to our filename or IO object.

        Returns:
            bool: Indicating success or failure
        """

        _, temp_filename = mkstemp()

        try:
            # Copy the file using `cp` such that we support sudo/su
            status, output = self.run_shell_command(
                StringCommand("cp", remote_filename, temp_filename),
                print_output=print_output,
                print_input=print_input,
                **arguments,
            )

            if not status:
                raise IOError(output.stderr)

            # Load our file or IO object and write it to the temporary file
            with open(temp_filename, "rb") as temp_f:
                with get_file_io(filename_or_io, "wb") as file_io:
                    data_bytes: bytes

                    data = temp_f.read()
                    if isinstance(data, str):
                        data_bytes = data.encode()
                    else:
                        data_bytes = data

                    file_io.write(data_bytes)
        finally:
            os.remove(temp_filename)

        if print_output:
            click.echo(
                "{0}file copied: {1}".format(self.host.print_prefix, remote_filename),
                err=True,
            )

        return True

    @override
    def check_can_rsync(self) -> None:
        if not which("rsync"):
            raise NotImplementedError("The `rsync` binary is not available on this system.")

    @override
    def rsync(
        self,
        src,
        dest,
        flags,
        print_output: bool = False,
        print_input: bool = False,
        **arguments,
    ) -> bool:
        status, output = self.run_shell_command(
            StringCommand("rsync", " ".join(flags), src, dest),
            print_output=print_output,
            print_input=print_input,
            **arguments,
        )

        if not status:
            raise IOError(output.stderr)

        return True



from collections import defaultdict
from os import path
from unittest import TestCase
from unittest.mock import mock_open, patch

import pyinfra
from pyinfra.api import (
    BaseStateCallback,
    Config,
    FileDownloadCommand,
    FileUploadCommand,
    OperationError,
    OperationValueError,
    State,
    StringCommand,
)
from pyinfra.api.connect import connect_all, disconnect_all
from pyinfra.api.exceptions import PyinfraError
from pyinfra.api.operation import OperationMeta, add_op
from pyinfra.api.operations import run_ops
from pyinfra.api.state import StateOperationMeta, StateStage
from pyinfra.connectors.util import CommandOutput, OutputLine
from pyinfra.context import ctx_host, ctx_state
from pyinfra.operations import files, python, server

from ..paramiko_util import FakeBuffer, FakeChannel, PatchSSHTestCase
from ..util import make_inventory


class TestOperationMeta(TestCase):
    def test_operation_meta_repr_no_change(self):
        op_meta = OperationMeta("hash", False)
        assert repr(op_meta) == "OperationMeta(executed=False, maybeChange=False, hash=hash)"

    def test_operation_meta_repr_changes(self):
        op_meta = OperationMeta("hash", True)
        assert repr(op_meta) == "OperationMeta(executed=False, maybeChange=True, hash=hash)"


class TestOperationsApi(PatchSSHTestCase):
    def test_op(self):
        inventory = make_inventory()
        somehost = inventory.get_host("somehost")
        anotherhost = inventory.get_host("anotherhost")

        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        state.add_callback_handler(BaseStateCallback())

        # Enable printing on this test to catch any exceptions in the formatting
        state.print_output = True
        state.print_input = True
        state.print_fact_info = True
        state.print_noop_info = True

        connect_all(state)

        add_op(
            state,
            files.file,
            "/var/log/pyinfra.log",
            user="pyinfra",
            group="pyinfra",
            mode="644",
            create_remote_dir=False,
            _sudo=True,
            _sudo_user="test_sudo",
            _su_user="test_su",
            _ignore_errors=True,
            _env={
                "TEST": "what",
            },
        )

        op_order = state.get_op_order()

        # Ensure we have an op
        assert len(op_order) == 1

        first_op_hash = op_order[0]

        # Ensure the op name
        assert state.op_meta[first_op_hash].names == {"files.file"}

        # Ensure the global kwargs (same for both hosts)
        somehost_global_arguments = state.ops[somehost][first_op_hash].global_arguments
        assert somehost_global_arguments["_sudo"] is True
        assert somehost_global_arguments["_sudo_user"] == "test_sudo"
        assert somehost_global_arguments["_su_user"] == "test_su"
        assert somehost_global_arguments["_ignore_errors"] is True

        anotherhost_global_arguments = state.ops[anotherhost][first_op_hash].global_arguments
        assert anotherhost_global_arguments["_sudo"] is True
        assert anotherhost_global_arguments["_sudo_user"] == "test_sudo"
        assert anotherhost_global_arguments["_su_user"] == "test_su"
        assert anotherhost_global_arguments["_ignore_errors"] is True

        # Ensure run ops works
        run_ops(state)

        # Ensure the commands
        assert state.ops[somehost][first_op_hash].operation_meta._commands == [
            StringCommand("touch /var/log/pyinfra.log"),
            StringCommand("chmod 644 /var/log/pyinfra.log"),
            StringCommand("chown pyinfra:pyinfra /var/log/pyinfra.log"),
        ]

        # Ensure ops completed OK
        assert state.results[somehost].success_ops == 1
        assert state.results[somehost].ops == 1
        assert state.results[anotherhost].success_ops == 1
        assert state.results[anotherhost].ops == 1

        # And w/o errors
        assert state.results[somehost].error_ops == 0
        assert state.results[anotherhost].error_ops == 0

        disconnect_all(state)

    @patch("pyinfra.api.util.open", mock_open(read_data="test!"), create=True)
    @patch("pyinfra.operations.files.os.path.isfile", lambda *args, **kwargs: True)
    def test_file_upload_op(self):
        inventory = make_inventory()

        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        # Test normal
        add_op(
            state,
            files.put,
            name="First op name",
            src="files/file.txt",
            dest="/home/vagrant/file.txt",
        )

        # And with sudo
        add_op(
            state,
            files.put,
            src="files/file.txt",
            dest="/home/vagrant/file.txt",
            _sudo=True,
            _sudo_user="pyinfra",
        )

        # And with su
        add_op(
            state,
            files.put,
            src="files/file.txt",
            dest="/home/vagrant/file.txt",
            _sudo=True,
            _su_user="pyinfra",
        )

        op_order = state.get_op_order()

        # Ensure we have all ops
        assert len(op_order) == 3

        first_op_hash = op_order[0]
        second_op_hash = op_order[1]

        # Ensure first op is the right one
        assert state.op_meta[first_op_hash].names == {"First op name"}

        somehost = inventory.get_host("somehost")
        anotherhost = inventory.get_host("anotherhost")

        # Ensure second op has sudo/sudo_user
        assert state.ops[somehost][second_op_hash].global_arguments["_sudo"] is True
        assert state.ops[somehost][second_op_hash].global_arguments["_sudo_user"] == "pyinfra"

        # Ensure third has su_user
        assert state.ops[somehost][op_order[2]].global_arguments["_su_user"] == "pyinfra"

        # Check run ops works
        run_ops(state)

        # Ensure first op used the right (upload) command
        assert state.ops[somehost][first_op_hash].operation_meta._commands == [
            StringCommand("mkdir -p /home/vagrant"),
            FileUploadCommand("files/file.txt", "/home/vagrant/file.txt"),
        ]

        # Ensure ops completed OK
        assert state.results[somehost].success_ops == 3
        assert state.results[somehost].ops == 3
        assert state.results[anotherhost].success_ops == 3
        assert state.results[anotherhost].ops == 3

        # And w/o errors
        assert state.results[somehost].error_ops == 0
        assert state.results[anotherhost].error_ops == 0

    def test_file_download_op(self):
        inventory = make_inventory()

        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        with patch("pyinfra.operations.files.os.path.isfile", lambda *args, **kwargs: True):
            add_op(
                state,
                files.get,
                name="First op name",
                src="/home/vagrant/file.txt",
                dest="files/file.txt",
            )

        op_order = state.get_op_order()

        assert len(op_order) == 1

        first_op_hash = op_order[0]
        assert state.op_meta[first_op_hash].names == {"First op name"}

        somehost = inventory.get_host("somehost")
        anotherhost = inventory.get_host("anotherhost")

        with patch("pyinfra.api.util.open", mock_open(read_data="test!"), create=True):
            run_ops(state)

        # Ensure first op has the right (upload) command
        assert state.ops[somehost][first_op_hash].operation_meta._commands == [
            FileDownloadCommand("/home/vagrant/file.txt", "files/file.txt"),
        ]

        assert state.results[somehost].success_ops == 1
        assert state.results[somehost].ops == 1
        assert state.results[anotherhost].success_ops == 1
        assert state.results[anotherhost].ops == 1
        assert state.results[somehost].error_ops == 0
        assert state.results[anotherhost].error_ops == 0

    def test_function_call_op(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        is_called = []

        def mocked_function(*args, **kwargs):
            is_called.append(True)
            return None

        # Add op to both hosts
        add_op(state, python.call, mocked_function)

        # Ensure there is one op
        assert len(state.get_op_order()) == 1

        run_ops(state)

        assert is_called

    def test_run_once_serial_op(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        # Add a run once op
        add_op(state, server.shell, 'echo "hi"', _run_once=True, _serial=True)

        # Ensure it's added to op_order
        assert len(state.get_op_order()) == 1

        somehost = inventory.get_host("somehost")
        anotherhost = inventory.get_host("anotherhost")

        # Ensure between the two hosts we only run the one op
        assert len(state.ops[somehost]) + len(state.ops[anotherhost]) == 1

        # Check run works
        run_ops(state)

        assert (state.results[somehost].success_ops + state.results[anotherhost].success_ops) == 1

    @patch("pyinfra.connectors.ssh.SSHConnector.check_can_rsync", lambda _: True)
    def test_rsync_op(self):
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        add_op(state, files.rsync, "src", "dest", _sudo=True, _sudo_user="root")

        assert len(state.get_op_order()) == 1

        with patch("pyinfra.connectors.ssh.run_local_process") as fake_run_local_process:
            fake_run_local_process.return_value = 0, []
            run_ops(state)

        fake_run_local_process.assert_called_with(
            (
                "rsync -ax --delete --rsh "
                '"ssh -o BatchMode=yes -o \\"StrictHostKeyChecking=accept-new\\""'
                " --rsync-path 'sudo -u root rsync' src vagrant@somehost:dest"
            ),
            print_output=False,
            print_prefix=inventory.get_host("somehost").print_prefix,
        )

    @patch("pyinfra.connectors.ssh.SSHConnector.check_can_rsync", lambda _: True)
    def test_rsync_op_with_strict_host_key_checking_disabled(self):
        inventory = make_inventory(hosts=(("somehost", {"ssh_strict_host_key_checking": "no"}),))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        add_op(state, files.rsync, "src", "dest", _sudo=True, _sudo_user="root")

        assert len(state.get_op_order()) == 1

        with patch("pyinfra.connectors.ssh.run_local_process") as fake_run_local_process:
            fake_run_local_process.return_value = 0, []
            run_ops(state)

        fake_run_local_process.assert_called_with(
            (
                "rsync -ax --delete --rsh "
                '"ssh -o BatchMode=yes -o \\"StrictHostKeyChecking=no\\""'
                " --rsync-path 'sudo -u root rsync' src vagrant@somehost:dest"
            ),
            print_output=False,
            print_prefix=inventory.get_host("somehost").print_prefix,
        )

    @patch("pyinfra.connectors.ssh.SSHConnector.check_can_rsync", lambda _: True)
    def test_rsync_op_with_strict_host_key_checking_disabled_and_custom_config_file(self):
        inventory = make_inventory(
            hosts=(
                (
                    "somehost",
                    {
                        "ssh_strict_host_key_checking": "no",
                        "ssh_config_file": "/home/me/ssh_test_config",
                    },
                ),
            )
        )
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        add_op(state, files.rsync, "src", "dest", _sudo=True, _sudo_user="root")

        assert len(state.get_op_order()) == 1

        with patch("pyinfra.connectors.ssh.run_local_process") as fake_run_local_process:
            fake_run_local_process.return_value = 0, []
            run_ops(state)

        fake_run_local_process.assert_called_with(
            (
                "rsync -ax --delete --rsh "
                '"ssh -o BatchMode=yes '
                '-o \\"StrictHostKeyChecking=no\\" -F /home/me/ssh_test_config"'
                " --rsync-path 'sudo -u root rsync' src vagrant@somehost:dest"
            ),
            print_output=False,
            print_prefix=inventory.get_host("somehost").print_prefix,
        )

    @patch("pyinfra.connectors.ssh.SSHConnector.check_can_rsync", lambda _: True)
    def test_rsync_op_with_sanitized_custom_config_file(self):
        inventory = make_inventory(
            hosts=(("somehost", {"ssh_config_file": "/home/me/ssh_test_config && echo hi"}),)
        )
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        add_op(state, files.rsync, "src", "dest", _sudo=True, _sudo_user="root")

        assert len(state.get_op_order()) == 1

        with patch("pyinfra.connectors.ssh.run_local_process") as fake_run_local_process:
            fake_run_local_process.return_value = 0, []
            run_ops(state)

        fake_run_local_process.assert_called_with(
            (
                "rsync -ax --delete --rsh "
                '"ssh -o BatchMode=yes -o \\"StrictHostKeyChecking=accept-new\\" '
                "-F '/home/me/ssh_test_config && echo hi'\""
                " --rsync-path 'sudo -u root rsync' src vagrant@somehost:dest"
            ),
            print_output=False,
            print_prefix=inventory.get_host("somehost").print_prefix,
        )

    def test_rsync_op_failure(self):
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        with patch("pyinfra.connectors.ssh.which", lambda x: None):
            with self.assertRaises(OperationError) as context:
                add_op(state, files.rsync, "src", "dest")

        assert context.exception.args[0] == "The `rsync` binary is not available on this system."

    def test_op_cannot_change_execution_kwargs(self):
        inventory = make_inventory()

        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare

        class NoSetDefaultDict(defaultdict):
            def setdefault(self, key, _):
                return self[key]

        op_meta_item = StateOperationMeta(tuple())
        op_meta_item.global_arguments = {"_serial": True}
        state.op_meta = NoSetDefaultDict(lambda: op_meta_item)

        connect_all(state)

        with self.assertRaises(OperationValueError) as context:
            add_op(state, files.file, "/var/log/pyinfra.log", _serial=False)

        assert context.exception.args[0] == "Cannot have different values for `_serial`."


class TestNestedOperationsApi(PatchSSHTestCase):
    def test_nested_op_api(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare

        connect_all(state)

        somehost = inventory.get_host("somehost")

        pyinfra.is_cli = True

        with ctx_state.use(state):
            with ctx_host.use(somehost):
                try:
                    outer_result = server.shell(commands="echo outer")
                    assert outer_result._combined_output is None

                    def callback():
                        inner_result = server.shell(commands="echo inner")
                        assert inner_result._combined_output is not None

                    python.call(function=callback)

                    assert len(state.get_op_order()) == 2

                    run_ops(state)

                    assert len(state.get_op_order()) == 3
                    assert state.results[somehost].success_ops == 3
                    assert outer_result._combined_output is not None

                    disconnect_all(state)
                finally:
                    pyinfra.is_cli = False

    def test_nested_op_failure_propagates_to_host(self):
        """
        Test that when an operation inside python.call fails,
        the host is properly marked as failed.
        Regression test for issue #1190.
        """
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        somehost = inventory.get_host("somehost")
        pyinfra.is_cli = True

        try:
            with ctx_state.use(state):
                with ctx_host.use(somehost):

                    def callback_with_failing_op():
                        server.shell(commands="/bin/false")

                    python.call(function=callback_with_failing_op)

            with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
                fake_channel = FakeChannel(1)
                fake_run_command.return_value = (False, FakeBuffer("", fake_channel))

                with self.assertRaises(PyinfraError) as e:
                    run_ops(state)

                self.assertEqual(e.exception.args[0], "No hosts remaining!")

            # Verify the host was marked as failed
            self.assertIn(somehost, state.failed_hosts)
            # Verify error_ops was incremented
            self.assertEqual(state.results[somehost].error_ops, 2)  # 1xcallbcak, 1xnested op
            self.assertEqual(state.results[somehost].success_ops, 0)
        finally:
            pyinfra.is_cli = False
            disconnect_all(state)

    def test_nested_op_failure_with_ignore_errors(self):
        """
        Test that _ignore_errors=True on python.call properly ignores
        errors from nested operations.
        """
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        somehost = inventory.get_host("somehost")
        pyinfra.is_cli = True

        try:
            with ctx_state.use(state):
                with ctx_host.use(somehost):

                    def callback_with_failing_op():
                        server.shell(commands="/bin/false")

                    python.call(function=callback_with_failing_op, _ignore_errors=True)

            with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
                fake_channel = FakeChannel(1)
                fake_run_command.return_value = (False, FakeBuffer("", fake_channel))

                # Should NOT raise - error is ignored
                run_ops(state)

            # Host should NOT be in failed_hosts
            self.assertNotIn(somehost, state.failed_hosts)
            # Error should be tracked as ignored
            self.assertEqual(state.results[somehost].ignored_error_ops, 1)
            self.assertEqual(state.results[somehost].success_ops, 0)
        finally:
            pyinfra.is_cli = False
            disconnect_all(state)

    def test_nested_op_partial_failure(self):
        """
        Test that when a callback has multiple operations and one fails,
        the failure is properly tracked.
        """
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        somehost = inventory.get_host("somehost")
        pyinfra.is_cli = True

        try:
            with ctx_state.use(state):
                with ctx_host.use(somehost):

                    def callback_with_multiple_ops():
                        server.shell(commands="echo success1")
                        server.shell(commands="/bin/false")  # This one fails
                        server.shell(commands="echo success2")

                    python.call(function=callback_with_multiple_ops)

            call_count = 0

            def side_effect(*args, **kwargs):
                nonlocal call_count
                call_count += 1
                if call_count == 2:  # Second command fails
                    return (False, FakeBuffer("", FakeChannel(1)))
                return (True, FakeBuffer("output", FakeChannel(0)))

            with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
                fake_run_command.side_effect = side_effect

                with self.assertRaises(PyinfraError):
                    run_ops(state)

            # Host should be marked as failed
            self.assertIn(somehost, state.failed_hosts)
            # Should have 1 success before failure, then 1 error
            self.assertEqual(state.results[somehost].success_ops, 1)
            self.assertEqual(state.results[somehost].error_ops, 2)
        finally:
            pyinfra.is_cli = False
            disconnect_all(state)

    def test_deeply_nested_op_failure(self):
        """
        Test that failures in deeply nested operations (callbacks within callbacks)
        are properly propagated.
        """
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        somehost = inventory.get_host("somehost")
        pyinfra.is_cli = True

        try:
            with ctx_state.use(state):
                with ctx_host.use(somehost):

                    def inner_callback():
                        server.shell(commands="/bin/false")

                    def outer_callback():
                        server.shell(commands="echo outer")
                        python.call(function=inner_callback)

                    python.call(function=outer_callback)

            call_count = 0

            def side_effect(*args, **kwargs):
                nonlocal call_count
                call_count += 1
                if call_count == 2:  # Inner shell fails
                    return (False, FakeBuffer("", FakeChannel(1)))
                return (True, FakeBuffer("output", FakeChannel(0)))

            with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
                fake_run_command.side_effect = side_effect

                with self.assertRaises(PyinfraError):
                    run_ops(state)

            # Host should be marked as failed
            self.assertIn(somehost, state.failed_hosts)
            self.assertGreater(state.results[somehost].error_ops, 0)
        finally:
            pyinfra.is_cli = False
            disconnect_all(state)


class TestOperationFailures(PatchSSHTestCase):
    def test_full_op_fail(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        add_op(state, server.shell, 'echo "hi"')

        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
            fake_channel = FakeChannel(1)
            fake_run_command.return_value = (
                False,
                FakeBuffer("", fake_channel),
            )

            with self.assertRaises(PyinfraError) as e:
                run_ops(state)

            assert e.exception.args[0] == "No hosts remaining!"

            somehost = inventory.get_host("somehost")

            # Ensure the op was not flagged as success
            assert state.results[somehost].success_ops == 0
            # And was flagged asn an error
            assert state.results[somehost].error_ops == 1

    def test_ignore_errors_op_fail(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        add_op(state, server.shell, 'echo "hi"', _ignore_errors=True)

        with patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command") as fake_run_command:
            fake_channel = FakeChannel(1)
            fake_run_command.return_value = (
                False,
                FakeBuffer("", fake_channel),
            )

            # This should run OK
            run_ops(state)

        somehost = inventory.get_host("somehost")

        # Ensure the op was added to results
        assert state.results[somehost].ops == 1
        assert state.results[somehost].ignored_error_ops == 1
        # But not as a success
        assert state.results[somehost].success_ops == 0


class TestOperationOrdering(PatchSSHTestCase):
    # In CLI mode, pyinfra uses *line numbers* to order operations as defined by
    # the user. This makes reasoning about user-written deploys simple and easy
    # to understand.
    def test_cli_op_line_numbers(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        state.current_deploy_filename = __file__

        pyinfra.is_cli = True

        with ctx_state.use(state):
            # Add op to both hosts
            for name in ("anotherhost", "somehost"):
                with ctx_host.use(inventory.get_host(name)):
                    server.shell("echo hi")  # note this is called twice but on *the same line*

            # Add op to just the second host - using the context modules such that
            # it replicates a deploy file.
            ctx_host.set(inventory.get_host("anotherhost"))
            first_context_hash = server.user("anotherhost_user")._hash

            # Add op to just the first host - using the context modules such that
            # it replicates a deploy file.
            ctx_host.set(inventory.get_host("somehost"))
            second_context_hash = server.user("somehost_user")._hash

            ctx_state.reset()
            ctx_host.reset()

            pyinfra.is_cli = False

            print(state.ops)
            # Ensure there are two ops
            op_order = state.get_op_order()
            assert len(op_order) == 3

            # And that the two ops above were called in the expected order
            assert op_order[1] == first_context_hash
            assert op_order[2] == second_context_hash

            # Ensure somehost has two ops and anotherhost only has the one
            assert len(state.ops[inventory.get_host("somehost")]) == 2
            assert len(state.ops[inventory.get_host("anotherhost")]) == 2

    # In API mode, pyinfra *overrides* the line numbers such that whenever an
    # operation or deploy is added it is simply appended. This makes sense as
    # the user writing the API calls has full control over execution order.
    def test_api_op_line_numbers(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        another_host = inventory.get_host("anotherhost")

        def add_another_op():
            return add_op(state, server.shell, "echo second-op")[another_host]._hash

        first_op_hash = add_op(state, server.shell, "echo first-op")[another_host]._hash
        second_op_hash = add_another_op()  # note `add_op` will be called on an earlier line

        op_order = state.get_op_order()
        assert len(op_order) == 2

        assert op_order[0] == first_op_hash
        assert op_order[1] == second_op_hash


class TestOperationRetry(PatchSSHTestCase):
    """
    Tests for the retry functionality in operations.
    """

    @patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command")
    def test_basic_retry_behavior(self, fake_run_command):
        """
        Test that operations retry the correct number of times on failure.
        """
        # Create inventory with just one host to simplify testing
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        # Add operation with retry settings
        add_op(
            state,
            server.shell,
            'echo "testing retries"',
            _retries=2,
            _retry_delay=0.1,  # Use small delay for tests
        )

        # Track how many times run_shell_command was called
        call_count = 0

        # First call fails, second succeeds
        def side_effect(*args, **kwargs):
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                # First call fails
                fake_channel = FakeChannel(1)
                return (False, FakeBuffer("", fake_channel))
            else:
                # Second call succeeds
                fake_channel = FakeChannel(0)
                return (True, FakeBuffer("success", fake_channel))

        fake_run_command.side_effect = side_effect

        # Run the operation
        run_ops(state)

        # Check that run_shell_command was called twice (original + 1 retry)
        self.assertEqual(call_count, 2)

        # Verify results
        somehost = inventory.get_host("somehost")

        # Operation should be successful (because the retry succeeded)
        self.assertEqual(state.results[somehost].success_ops, 1)
        self.assertEqual(state.results[somehost].error_ops, 0)

        # Get the operation hash
        op_hash = state.get_op_order()[0]

        # Check retry info in OperationMeta
        op_meta = state.ops[somehost][op_hash].operation_meta
        self.assertEqual(op_meta.retry_attempts, 1)
        self.assertEqual(op_meta.max_retries, 2)
        self.assertTrue(op_meta.was_retried)
        self.assertTrue(op_meta.retry_succeeded)

    @patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command")
    def test_retry_max_attempts_failure(self, fake_run_command):
        """
        Test that operations stop retrying after max attempts and report failure.
        """
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        # Add operation with retry settings
        add_op(
            state,
            server.shell,
            'echo "testing max retries"',
            _retries=2,
            _retry_delay=0.1,
        )

        # Make all attempts fail
        fake_channel = FakeChannel(1)
        fake_run_command.return_value = (False, FakeBuffer("", fake_channel))

        # This should fail after all retries
        with self.assertRaises(PyinfraError) as e:
            run_ops(state)

        self.assertEqual(e.exception.args[0], "No hosts remaining!")

        # Check that run_shell_command was called the right number of times (1 original + 2 retries)
        self.assertEqual(fake_run_command.call_count, 3)

        somehost = inventory.get_host("somehost")

        # Operation should be marked as error
        self.assertEqual(state.results[somehost].success_ops, 0)
        self.assertEqual(state.results[somehost].error_ops, 1)

        # Get the operation hash
        op_hash = state.get_op_order()[0]

        # Check retry info
        op_meta = state.ops[somehost][op_hash].operation_meta
        self.assertEqual(op_meta.retry_attempts, 2)
        self.assertEqual(op_meta.max_retries, 2)
        self.assertTrue(op_meta.was_retried)
        self.assertFalse(op_meta.retry_succeeded)

    @patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command")
    @patch("time.sleep")
    def test_retry_until_condition(self, fake_sleep, fake_run_command):
        """
        Test that operations retry based on the retry_until callable condition.
        """
        # Setup inventory and state using the utility function
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        # Create a counter to track retry_until calls
        call_counter = [0]

        # Create a retry_until function that returns True (retry) for first two calls
        def retry_until_func(output_data):
            call_counter[0] += 1
            return call_counter[0] < 3  # Retry twice, then stop

        # Add operation with retry_until
        add_op(
            state,
            server.shell,
            'echo "test retry_until"',
            _retries=3,
            _retry_delay=0.1,
            _retry_until=retry_until_func,
        )

        # Set up fake command execution - always succeed but with proper output format
        # Use the existing FakeBuffer/FakeChannel from test utils

        # First two calls trigger retry_until, third doesn't
        def command_side_effect(*args, **kwargs):
            # Create proper CommandOutput for the retry_until function to process
            lines = [OutputLine("stdout", "test output"), OutputLine("stderr", "no errors")]
            return True, CommandOutput(lines)

        fake_run_command.side_effect = command_side_effect

        # Run the operations
        run_ops(state)

        # The command should be called 3 times total (initial + 2 retries)
        self.assertEqual(fake_run_command.call_count, 3)

        # The retry_until function should be called 3 times
        self.assertEqual(call_counter[0], 3)

        # Get the operation metadata to check retry info
        somehost = inventory.get_host("somehost")
        op_hash = state.get_op_order()[0]
        op_meta = state.ops[somehost][op_hash].operation_meta

        # Check retry metadata
        self.assertEqual(op_meta.retry_attempts, 2)
        self.assertEqual(op_meta.max_retries, 3)
        self.assertTrue(op_meta.was_retried)
        self.assertTrue(op_meta.retry_succeeded)

    @patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command")
    @patch("time.sleep")
    def test_retry_delay(self, fake_sleep, fake_run_command):
        """
        Test that retry delay is properly applied between attempts.
        """
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        retry_delay = 5

        # Add operation with retry settings
        add_op(
            state,
            server.shell,
            'echo "testing retry delay"',
            _retries=2,
            _retry_delay=retry_delay,
        )

        # Make first call fail, second succeed
        call_count = 0

        def side_effect(*args, **kwargs):
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                fake_channel = FakeChannel(1)
                return (False, FakeBuffer("", fake_channel))
            else:
                fake_channel = FakeChannel(0)
                return (True, FakeBuffer("", fake_channel))

        fake_run_command.side_effect = side_effect

        # Run the operation
        run_ops(state)

        # Check that sleep was called with the correct delay
        fake_sleep.assert_called_once_with(retry_delay)

    @patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command")
    @patch("time.sleep")
    def test_retry_until_with_error_handling(self, fake_sleep, fake_run_command):
        """
        Test that operations handle errors in retry_until functions gracefully.
        """
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        # Create a retry_until function that raises an exception
        def failing_retry_until_func(output_data):
            raise ValueError("Test error in retry_until function")

        # Add operation with failing retry_until
        add_op(
            state,
            server.shell,
            'echo "test failing retry_until"',
            _retries=2,
            _retry_delay=0.1,
            _retry_until=failing_retry_until_func,
        )

        # Set up fake command execution

        def command_side_effect(*args, **kwargs):
            lines = [OutputLine("stdout", "test output"), OutputLine("stderr", "no errors")]
            return True, CommandOutput(lines)

        fake_run_command.side_effect = command_side_effect

        # Run the operations - should succeed despite retry_until error
        run_ops(state)

        # The command should be called only once (no retries due to error)
        self.assertEqual(fake_run_command.call_count, 1)

        # Verify operation completed successfully
        somehost = inventory.get_host("somehost")
        self.assertEqual(state.results[somehost].success_ops, 1)
        self.assertEqual(state.results[somehost].error_ops, 0)

    @patch("pyinfra.connectors.ssh.SSHConnector.run_shell_command")
    @patch("time.sleep")
    def test_retry_until_with_complex_output_parsing(self, fake_sleep, fake_run_command):
        """
        Test retry_until with complex output parsing scenarios.
        """
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        state.current_stage = StateStage.Prepare
        connect_all(state)

        # Track what output we've seen
        outputs_seen = []

        def complex_retry_until_func(output_data):
            # Store the output data for verification
            outputs_seen.append(output_data)

            # Check for specific patterns in stdout
            stdout_text = " ".join(output_data["stdout_lines"])

            # Continue retrying until we see "READY" in stdout
            return "READY" not in stdout_text

        # Add operation with complex retry_until
        add_op(
            state,
            server.shell,
            'echo "service status check"',
            _retries=3,
            _retry_delay=0.1,
            _retry_until=complex_retry_until_func,
        )

        # Set up fake command execution with changing output

        call_count = 0

        def command_side_effect(*args, **kwargs):
            nonlocal call_count
            call_count += 1

            if call_count == 1:
                lines = [
                    OutputLine("stdout", "Service starting..."),
                    OutputLine("stderr", "Loading config"),
                ]
            elif call_count == 2:
                lines = [
                    OutputLine("stdout", "Service initializing..."),
                    OutputLine("stderr", "Connecting to database"),
                ]
            else:  # call_count == 3
                lines = [
                    OutputLine("stdout", "Service READY"),
                    OutputLine("stderr", "All systems operational"),
                ]

            return True, CommandOutput(lines)

        fake_run_command.side_effect = command_side_effect

        # Run the operations
        run_ops(state)

        # The command should be called 3 times
        self.assertEqual(fake_run_command.call_count, 3)

        # Verify retry_until was called 3 times with correct data
        self.assertEqual(len(outputs_seen), 3)

        # Check the output data structure
        for output_data in outputs_seen:
            self.assertIn("stdout_lines", output_data)
            self.assertIn("stderr_lines", output_data)
            self.assertIn("commands", output_data)
            self.assertIn("executed_commands", output_data)
            self.assertIn("host", output_data)
            self.assertIn("operation", output_data)

        # Verify operation metadata
        somehost = inventory.get_host("somehost")
        op_hash = state.get_op_order()[0]
        op_meta = state.ops[somehost][op_hash].operation_meta

        self.assertEqual(op_meta.retry_attempts, 2)
        self.assertEqual(op_meta.max_retries, 3)
        self.assertTrue(op_meta.was_retried)
        self.assertTrue(op_meta.retry_succeeded)


this_filename = path.join("tests", "test_api", "test_api_operations.py")



"""
This file as originally part of the "sshuserclient" pypi package. The GitHub
source has now vanished (https://github.com/tobald/sshuserclient).
"""

from os import path

from gevent.lock import BoundedSemaphore
from paramiko import (
    HostKeys,
    MissingHostKeyPolicy,
    ProxyCommand,
    SSHClient as ParamikoClient,
    SSHException,
)
from paramiko.agent import AgentRequestHandler
from paramiko.hostkeys import HostKeyEntry
from typing_extensions import override

from pyinfra import logger
from pyinfra.api.util import memoize

from .config import SSHConfig

HOST_KEYS_LOCK = BoundedSemaphore()


class StrictPolicy(MissingHostKeyPolicy):
    @override
    def missing_host_key(self, client, hostname, key):
        logger.error("No host key for %s found in known_hosts", hostname)
        raise SSHException(f"StrictPolicy: No host key for {hostname} found in known_hosts")


def append_hostkey(client, hostname, key):
    """Append hostname to the clients host_keys_file"""

    if client._host_keys_filename is None:
        logger.warning("No host keys filename, not saving key for: %s", hostname)
        return

    with HOST_KEYS_LOCK:
        # The paramiko client saves host keys incorrectly whereas the host keys object does
        # this correctly, so use that with the client filename variable.
        # See: https://github.com/paramiko/paramiko/pull/1989
        host_key_entry = HostKeyEntry([hostname], key)
        if host_key_entry is None:
            raise SSHException(
                "Append Hostkey: Failed to parse host {0}, could not append to hostfile".format(
                    hostname
                ),
            )
        with open(client._host_keys_filename, "a") as host_keys_file:
            hk_entry = host_key_entry.to_line()
            if hk_entry is None:
                raise SSHException(f"Append Hostkey: Failed to append hostkey ({host_key_entry})")

            host_keys_file.write(hk_entry)


class AcceptNewPolicy(MissingHostKeyPolicy):
    @override
    def missing_host_key(self, client, hostname, key):
        logger.warning(
            (
                f"No host key for {hostname} found in known_hosts, "
                "accepting & adding to host keys file"
            ),
        )

        append_hostkey(client, hostname, key)
        logger.warning("Added host key for %s to known_hosts", hostname)


class AskPolicy(MissingHostKeyPolicy):
    @override
    def missing_host_key(self, client, hostname, key):
        should_continue = input(
            f"No host key for {hostname} found in known_hosts, do you want to continue [y/n] ",
        )
        if should_continue.lower() != "y":
            raise SSHException(
                f"AskPolicy: No host key for {hostname} found in known_hosts",
            )
        append_hostkey(client, hostname, key)
        logger.warning("Added host key for %s to known_hosts", hostname)
        return


class WarningPolicy(MissingHostKeyPolicy):
    @override
    def missing_host_key(self, client, hostname, key):
        logger.warning("No host key for %s found in known_hosts", hostname)


def get_missing_host_key_policy(policy):
    if policy is None or policy == "ask":
        return AskPolicy()
    if policy == "no" or policy == "off":
        return WarningPolicy()
    if policy == "yes":
        return StrictPolicy()
    if policy == "accept-new":
        return AcceptNewPolicy()
    raise SSHException(f"Invalid value StrictHostKeyChecking={policy}")


@memoize
def get_ssh_config(user_config_file=None):
    logger.debug("Loading SSH config: %s", user_config_file)

    if user_config_file is None:
        user_config_file = path.expanduser("~/.ssh/config")

    if path.exists(user_config_file):
        with open(user_config_file, encoding="utf-8") as f:
            ssh_config = SSHConfig()
            ssh_config.parse(f)
            return ssh_config


@memoize
def get_host_keys(filenames):
    """
    Load host keys from one or more files.

    Args:
        filenames: A tuple of filenames to load host keys from.
    """
    with HOST_KEYS_LOCK:
        host_keys = HostKeys()

        for filename in filenames:
            try:
                host_keys.load(filename)
            # When paramiko encounters a bad host keys line it sometimes bails the
            # entire load incorrectly.
            # See: https://github.com/paramiko/paramiko/pull/1990
            except Exception as e:
                logger.warning("Failed to load host keys from %s: %s", filename, e)

        return host_keys


class SSHClient(ParamikoClient):
    """
    An SSHClient which honors ssh_config and supports proxyjumping
    original idea at http://bitprophet.org/blog/2012/11/05/gateway-solutions/.
    """

    @override
    def connect(  # type: ignore[override]
        self,
        hostname,
        _pyinfra_ssh_forward_agent=None,
        _pyinfra_ssh_config_file=None,
        _pyinfra_ssh_known_hosts_file=None,
        _pyinfra_ssh_strict_host_key_checking=None,
        _pyinfra_ssh_paramiko_connect_kwargs=None,
        **kwargs,
    ):
        (
            hostname,
            config,
            forward_agent,
            missing_host_key_policy,
            host_keys_files,
            keep_alive,
        ) = self.parse_config(
            hostname,
            kwargs,
            ssh_config_file=_pyinfra_ssh_config_file,
            strict_host_key_checking=_pyinfra_ssh_strict_host_key_checking,
        )
        self.set_missing_host_key_policy(missing_host_key_policy)
        config.update(kwargs)

        if _pyinfra_ssh_known_hosts_file:
            host_keys_files = (path.expanduser(_pyinfra_ssh_known_hosts_file),)

        # Overwrite paramiko empty defaults with @memoize-d host keys object
        self._host_keys = get_host_keys(host_keys_files)
        # Use the first file for writing new host keys
        if len(host_keys_files) > 0:
            self._host_keys_filename = host_keys_files[0]

        if _pyinfra_ssh_paramiko_connect_kwargs:
            config.update(_pyinfra_ssh_paramiko_connect_kwargs)

        self._ssh_config = config
        super().connect(hostname, **config)

        if _pyinfra_ssh_forward_agent is not None:
            forward_agent = _pyinfra_ssh_forward_agent

        if keep_alive:
            transport = self.get_transport()
            assert transport is not None, "No transport"
            transport.set_keepalive(keep_alive)

        if forward_agent:
            transport = self.get_transport()
            assert transport is not None, "No transport"
            session = transport.open_session()
            AgentRequestHandler(session)

    def gateway(self, hostname, host_port, target, target_port):
        transport = self.get_transport()
        assert transport is not None, "No transport"
        return transport.open_channel(
            "direct-tcpip",
            (target, target_port),
            (hostname, host_port),
        )

    def parse_config(
        self,
        hostname,
        initial_cfg=None,
        ssh_config_file=None,
        strict_host_key_checking=None,
    ):
        cfg: dict = {"port": 22}
        cfg.update(initial_cfg or {})

        keep_alive = 0
        forward_agent = False
        missing_host_key_policy = get_missing_host_key_policy(strict_host_key_checking)
        host_keys_files = (path.expanduser("~/.ssh/known_hosts"),)

        ssh_config = get_ssh_config(ssh_config_file)
        if not ssh_config:
            return (
                hostname,
                cfg,
                forward_agent,
                missing_host_key_policy,
                host_keys_files,
                keep_alive,
            )

        host_config = ssh_config.lookup(hostname)
        forward_agent = host_config.get("forwardagent") == "yes"

        # If not overridden, apply any StrictHostKeyChecking
        if strict_host_key_checking is None and "stricthostkeychecking" in host_config:
            missing_host_key_policy = get_missing_host_key_policy(
                host_config["stricthostkeychecking"],
            )

        if "userknownhostsfile" in host_config:
            # OpenSSH supports multiple space-separated known hosts files
            host_keys_files = tuple(
                path.expanduser(f) for f in host_config["userknownhostsfile"].split()
            )

        if "hostname" in host_config:
            hostname = host_config["hostname"]

        if "user" in host_config:
            cfg["username"] = host_config["user"]

        if "identityfile" in host_config:
            cfg["key_filename"] = host_config["identityfile"]

        if "port" in host_config:
            cfg["port"] = int(host_config["port"])

        if "serveraliveinterval" in host_config:
            keep_alive = int(host_config["serveraliveinterval"])

        if "proxycommand" in host_config:
            cfg["sock"] = ProxyCommand(host_config["proxycommand"])

        elif "proxyjump" in host_config:
            hops = host_config["proxyjump"].split(",")
            sock = None

            for i, hop in enumerate(hops):
                hop_hostname, hop_config = self.derive_shorthand(ssh_config, hop)
                logger.debug("SSH ProxyJump through %s:%s", hop_hostname, hop_config["port"])

                c = SSHClient()
                c.connect(
                    hop_hostname, _pyinfra_ssh_config_file=ssh_config_file, sock=sock, **hop_config
                )

                if i == len(hops) - 1:
                    target = hostname
                    target_config = {"port": cfg["port"]}
                else:
                    target, target_config = self.derive_shorthand(ssh_config, hops[i + 1])

                sock = c.gateway(hostname, cfg["port"], target, target_config["port"])
            cfg["sock"] = sock

        return hostname, cfg, forward_agent, missing_host_key_policy, host_keys_files, keep_alive

    @staticmethod
    def derive_shorthand(ssh_config, host_string):
        shorthand_config = {}
        user_hostport = host_string.rsplit("@", 1)
        hostport = user_hostport.pop()
        user = user_hostport[0] if user_hostport and user_hostport[0] else None
        if user:
            shorthand_config["username"] = user

        # IPv6: can't reliably tell where addr ends and port begins, so don't
        # try (and don't bother adding special syntax either, user should avoid
        # this situation by using port=).
        if hostport.count(":") > 1:
            hostname = hostport
        # IPv4: can split on ':' reliably.
        else:
            host_port = hostport.rsplit(":", 1)
            hostname = host_port.pop(0) or None
            if host_port and host_port[0]:
                shorthand_config["port"] = int(host_port[0])

        base_config = ssh_config.lookup(hostname)

        config = {
            "port": base_config.get("port", 22),
            "username": base_config.get("user"),
        }
        config.update(shorthand_config)

        return hostname, config



"""
SSH integration tests - mostly checking for idempotency on file operations.
"""

import time

import pytest

DOCKER_IMAGE = (
    "linuxserver/openssh-server@sha256:"
    "1e33f77de1021dcd2e52a6bbc787082abfb591416a00e6a1422b83b3b426f09d"
)

PYINFRA_COMMAND = (
    "pyinfra -y -v localhost "
    "--ssh-port 2222 "
    "--ssh-user pyinfra "
    "--ssh-password password "
    "--data ssh_known_hosts_file=/dev/null "
    "--data ssh_strict_host_key_checking=no "
)


@pytest.fixture(autouse=True)
def run_docker_ssh_server(helpers):
    docker_container_id, _ = helpers.run(
        " ".join(
            [
                "docker run -d",
                "-p 2222:2222",
                "-e PASSWORD_ACCESS=true",
                "-e USER_PASSWORD=password",
                "-e USER_NAME=pyinfra",
                "-e SUDO_ACCESS=true",
                DOCKER_IMAGE,
            ],
        ),
    )
    time.sleep(10)  # allow openSSH to start
    try:
        yield docker_container_id
    finally:
        helpers.run(f"docker kill {docker_container_id}")


@pytest.mark.end_to_end
@pytest.mark.end_to_end_ssh
def test_e2e_ssh_sudo_password(helpers):
    helpers.run_check_output(
        f"{PYINFRA_COMMAND} server.shell echo _sudo=True _sudo_password=password",
        expected_lines=["localhost] Success"],
    )
    helpers.run_check_output(
        f"{PYINFRA_COMMAND} server.shell echo _sudo=True _sudo_password=wrongpassword",
        expected_lines=["localhost] Error"],
        expected_exit_code=1,
    )


@pytest.mark.end_to_end
@pytest.mark.end_to_end_ssh
def test_int_local_file_no_changes(helpers):
    helpers.run_check_output(  # first run = create the file
        f"{PYINFRA_COMMAND} files.file _testfile",
        expected_lines=["localhost] Success"],
    )

    helpers.run_check_output(  # second run = no changes
        f"{PYINFRA_COMMAND} files.file _testfile",
        expected_lines=["localhost] No changes"],
    )


@pytest.mark.end_to_end
@pytest.mark.end_to_end_ssh
def test_int_local_directory_no_changes(helpers):
    helpers.run_check_output(  # first run = create the directory
        f"{PYINFRA_COMMAND} files.directory _testdir",
        expected_lines=["localhost] Success"],
    )

    helpers.run_check_output(  # second run = no changes
        f"{PYINFRA_COMMAND} files.directory _testdir",
        expected_lines=["localhost] No changes"],
    )

    helpers.run_check_output(  # third run (remove) = remove directory
        f"{PYINFRA_COMMAND} files.directory _testdir present=False",
        expected_lines=["localhost] Success"],
    )

    helpers.run_check_output(  # fourth run (remove) = no chances
        f"{PYINFRA_COMMAND} files.directory _testdir present=False",
        expected_lines=["localhost] No changes"],
    )


@pytest.mark.end_to_end
@pytest.mark.end_to_end_ssh
def test_int_local_link_no_changes(helpers):
    helpers.run_check_output(  # first run = create the link
        f"{PYINFRA_COMMAND} files.link _testlink target=_testfile",
        expected_lines=["localhost] Success"],
    )

    helpers.run_check_output(  # second run = no changes
        f"{PYINFRA_COMMAND} files.link _testlink target=_testfile",
        expected_lines=["localhost] No changes"],
    )


@pytest.mark.end_to_end
@pytest.mark.end_to_end_ssh
def test_int_local_line_no_changes(helpers):
    helpers.run_check_output(  # first run = create the line
        f"{PYINFRA_COMMAND} files.line _testfile someline",
        expected_lines=["localhost] Success"],
    )

    helpers.run_check_output(  # second run = no changes
        f"{PYINFRA_COMMAND} files.line _testfile someline",
        expected_lines=["localhost] No changes"],
    )

    helpers.run_check_output(  # replace the line
        f"{PYINFRA_COMMAND} files.line _testfile someline replace=anotherline",
        expected_lines=["localhost] Success"],
    )

    helpers.run_check_output(  # second run replace the line = no changes
        f"{PYINFRA_COMMAND} files.line _testfile someline replace=anotherline",
        expected_lines=["localhost] No changes"],
    )



try:
    from importlib_metadata import entry_points
except ImportError:
    from importlib.metadata import entry_points  # type: ignore[assignment]


def _load_connector(entrypoint):
    return entrypoint.load()


def get_all_connectors():
    return {
        entrypoint.name: _load_connector(entrypoint)
        for entrypoint in entry_points(group="pyinfra.connectors")
    }


def get_execution_connectors():
    return {
        connector: connector_mod
        for connector, connector_mod in get_all_connectors().items()
        if connector_mod.handles_execution
    }


def get_execution_connector(name):
    return get_execution_connectors()[name]



#!/usr/bin/env python

import sys
from glob import glob
from importlib import import_module
from inspect import getfullargspec, getmembers, isclass
from os import makedirs, path
from types import FunctionType, MethodType

from pyinfra.api.facts import FactBase, ShortFactBase

sys.path.append(".")
from docs.utils import format_doc_line, title_line  # noqa: E402


def build_facts_docs():
    this_dir = path.dirname(path.realpath(__file__))
    docs_dir = path.abspath(path.join(this_dir, "..", "docs"))
    facts_dir = path.join(this_dir, "..", "src", "pyinfra", "facts", "*.py")

    makedirs(path.join(docs_dir, "facts"), exist_ok=True)

    fact_module_names = [
        path.basename(name)[:-3] for name in glob(facts_dir) if not name.endswith("__init__.py")
    ]

    for module_name in sorted(fact_module_names):
        lines = []
        print("--> Doing fact module: {0}".format(module_name))
        module = import_module("pyinfra.facts.{0}".format(module_name))

        full_title = "{0} Facts".format(module_name.title())
        lines.append(full_title)
        lines.append(title_line("-", full_title))
        lines.append("")

        if module.__doc__:
            lines.append(module.__doc__)

        if path.exists(path.join(this_dir, "..", "pyinfra", "operations", f"{module_name}.py")):
            lines.append(f"See also: :doc:`../operations/{module_name}`.")
            lines.append("")

        fact_classes = [
            (key, value)
            for key, value in getmembers(module)
            if (
                isclass(value)
                and (issubclass(value, FactBase) or issubclass(value, ShortFactBase))
                and value.__module__ == module.__name__
                and value is not FactBase
                and not value.__name__.endswith("Base")  # hacky!
            )
        ]

        for fact, cls in fact_classes:
            # FactClass -> fact_accessor on host object
            name = fact
            args_string_and_brackets = ""

            # Does this fact take args?
            command_attr = getattr(cls, "command", None)
            if isinstance(command_attr, (FunctionType, MethodType)):
                # Attach basic argspec to name
                # Note only supports facts with one arg as this is all that's
                # possible, will need to refactor to print properly in future.
                argspec = getfullargspec(command_attr)

                arg_defaults = (
                    [
                        "'{}'".format(arg) if isinstance(arg, str) else arg
                        for arg in argspec.defaults
                    ]
                    if argspec.defaults
                    else None
                )

                # Create a dict of arg name -> default
                defaults = (
                    dict(
                        zip(
                            argspec.args[-len(arg_defaults) :],
                            arg_defaults,
                        ),
                    )
                    if arg_defaults
                    else {}
                )

                if len(argspec.args):
                    args_string_and_brackets = ", {0}".format(
                        ", ".join(
                            ("{0}={1}".format(arg, defaults.get(arg)) if arg in defaults else arg)
                            for arg in argspec.args
                            if arg != "self"
                        ),
                    )

            lines.append(".. _facts:{0}.{1}:".format(module_name, name))
            lines.append("")

            title = ":code:`{0}.{1}`".format(module_name, name)
            lines.append(title)

            # Underline name with -'s for title
            lines.append(title_line("~", title))
            lines.append("")

            # Attach the code block
            lines.append(
                """
.. code:: python

    host.get_fact({1}{2})

""".strip().format(
                    module_name,
                    name,
                    args_string_and_brackets,
                ),
            )

            # Append any docstring
            doc = cls.__doc__
            if doc:
                lines.append("")
                lines.append(
                    "{0}".format(
                        "\n".join([format_doc_line(line) for line in doc.split("\n")]),
                    ).strip(),
                )

            lines.append("")
            lines.append("")

        # Write out the file
        module_filename = path.join(docs_dir, "facts", "{0}.rst".format(module_name))
        print("--> Writing {0}".format(module_filename))

        with open(module_filename, "w", encoding="utf-8") as outfile:
            outfile.write("\n".join(lines))


if __name__ == "__main__":
    print("### Building fact docs")
    build_facts_docs()



from __future__ import annotations

import re
import shlex

from typing_extensions import override

from pyinfra.api import FactBase

from .util.packaging import parse_packages

rpm_regex = r"^(\S+)\ (\S+)$"
rpm_query_format = "%{NAME} %{VERSION}-%{RELEASE}\\n"


class RpmPackages(FactBase):
    """
    Returns a dict of installed rpm packages:

    .. code:: python

        {
            "package_name": ["version"],
        }
    """

    @override
    def command(self) -> str:
        return "rpm --queryformat {0} -qa".format(shlex.quote(rpm_query_format))

    @override
    def requires_command(self) -> str:
        return "rpm"

    default = dict

    @override
    def process(self, output):
        return parse_packages(rpm_regex, output)


class RpmPackage(FactBase):
    """
    Returns information on a .rpm file:

    .. code:: python

        {
            "name": "my_package",
            "version": "1.0.0",
        }
    """

    @override
    def requires_command(self, package) -> str:
        return "rpm"

    @override
    def command(self, package) -> str:
        return (
            "rpm --queryformat {0} -q {1} || "
            "! test -e {1} || "
            "rpm --queryformat {0} -qp {1} 2> /dev/null"
        ).format(shlex.quote(rpm_query_format), shlex.quote(package))

    @override
    def process(self, output):
        for line in output:
            matches = re.match(rpm_regex, line)
            if matches:
                return {
                    "name": matches.group(1),
                    "version": matches.group(2),
                }


class RpmPackageProvides(FactBase):
    """
    Returns a list of packages that provide the specified capability (command, file, etc).
    """

    default = list

    @override
    def requires_command(self, *args, **kwargs) -> str:
        return "repoquery"

    @override
    def command(self, package):
        # Accept failure here (|| true) for invalid/unknown packages
        return "repoquery --queryformat {0} --whatprovides {1} || true".format(
            shlex.quote(rpm_query_format),
            shlex.quote(package),
        )

    @override
    def process(self, output):
        packages = []

        for line in output:
            matches = re.match(rpm_regex, line)
            if matches:
                packages.append(list(matches.groups()))

        return packages



from unittest import TestCase

from pyinfra import host, inventory
from pyinfra.api import Host, Inventory
from pyinfra.context import ctx_host, ctx_inventory


def _create_host(name: str = "host"):
    return Host(name, Inventory(([name], {})), None, None)


class TestHostContextObject(TestCase):
    def test_context_host_dir_matches_class(self):
        assert dir(Host) == dir(host)

    def test_context_host_eq(self):
        host_obj = _create_host()
        with ctx_host.use(host_obj):
            assert host == host_obj

    def test_context_host_repr(self):
        host_obj = _create_host("somehost")
        with ctx_host.use(host_obj):
            assert repr(host) == "ContextObject(Host):Host(somehost)"

    def test_context_host_str(self):
        host_obj = _create_host()
        with ctx_host.use(host_obj):
            assert str(host_obj) == "host"

    def test_context_host_attr(self):
        host_obj = _create_host()

        with self.assertRaises(AttributeError):
            host_obj.hello

        with ctx_host.use(host_obj):
            setattr(host, "hello", "world")
            assert host_obj.hello == host.hello

    def test_context_host_class_attr(self):
        host_obj = _create_host()

        with ctx_host.use(host_obj):
            assert ctx_host.isset() is True

            with self.assertRaises(AttributeError):
                host_obj.hello

            setattr(Host, "hello", "class_world")
            setattr(host_obj, "hello", "instance_world")

            assert host.hello == host.hello

        # Reset and check fallback to class variable
        assert ctx_host.isset() is False
        assert host.hello == "class_world"


class TestInventoryContextObject(TestCase):
    def test_context_inventory_len(self):
        inventory_obj = Inventory(("host", "anotherhost"))

        with ctx_inventory.use(inventory_obj):
            assert ctx_inventory.isset() is True
            assert len(inventory) == len(inventory_obj)

    def test_context_inventory_iter(self):
        inventory_obj = Inventory(("host", "anotherhost"))

        with ctx_inventory.use(inventory_obj):
            assert ctx_inventory.isset() is True
            assert list(iter(inventory)) == list(iter(inventory_obj))



# This deploy intentionally sleeps random intervals between operation calls, and
# different branches for each host. This is to ensure that preparing the operations
# in parallel always gives the same result, with the same set of facts for each host.

from random import randint
from time import sleep

from pyinfra import config, host, inventory
from pyinfra.facts.server import Arch, Os
from pyinfra.operations import python, server

config.SUDO = True

sleep(randint(1, 10) * 0.01)
server.shell(
    name="First main operation",
    commands="echo first_main_op",
)

# Create some conditional branches
if host.name == "somehost":
    sleep(randint(1, 10) * 0.01)
    server.shell(
        name="Second main somehost operation",
        commands="echo second_main_op",
    )
    # Give somehost two facts
    sleep(randint(1, 10) * 0.01)
    host.get_fact(Os)
    host.get_fact(Arch)

    assert config.SUDO is True

elif host.name == "anotherhost":
    config.SUDO = False

    sleep(randint(1, 10) * 0.01)
    server.shell(
        name="Second main anotherhost operation",
        commands="echo second_main_op",
    )

    assert config.SUDO is False

else:  # some other host
    host.get_fact(Os)

    # Load fact from a different host
    somehost = inventory.get_host("somehost")
    somehost.get_fact(Os)

    assert config.SUDO is True


def nested_op():
    sleep(randint(1, 10) * 0.01)
    server.shell(
        name="First nested operation",
        commands="echo first_nested_operation",
    )

    if host.name == "anotherhost":
        sleep(randint(1, 10) * 0.01)
        server.shell(
            name="Second nested anotherhost operation",
            commands="echo first_nested_operation",
        )

    if host.name == "somehost":
        server.shell(
            name="Second nested somehost operation",
            commands="echo first_nested_operation",
        )


python.call(name="Function call operation", function=nested_op)


sleep(randint(1, 10) * 0.01)
server.shell(
    name="Third main operation",
    commands="echo third_main_op",
)



from __future__ import annotations

from contextlib import contextmanager
from copy import copy
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)
from uuid import uuid4

import click
from typing_extensions import Unpack, override

from pyinfra import logger
from pyinfra.connectors.base import BaseConnector
from pyinfra.connectors.util import CommandOutput, remove_any_sudo_askpass_file

from .connectors import get_execution_connector
from .exceptions import ConnectError
from .facts import FactBase, ShortFactBase, get_fact
from .util import memoize, sha1_hash

if TYPE_CHECKING:
    from pyinfra.api.arguments import AllArguments
    from pyinfra.api.inventory import Inventory
    from pyinfra.api.state import State


def extract_callable_datas(
    datas: list[Union[Callable[..., Any], Any]],
) -> Generator[Any, Any, Any]:
    for data in datas:
        # Support for dynamic data, ie @deploy wrapped data defaults where
        # the data is stored on the state temporarily.
        if callable(data):
            data = data()
        yield data


class HostData:
    """
    Combines multiple AttrData's to search for attributes.
    """

    override_datas: dict[str, Any]

    def __init__(self, host: "Host", *datas):
        self.__dict__["host"] = host

        parsed_datas = list(datas)

        # Inject an empty override data so we can assign during deploy
        self.__dict__["override_datas"] = {}
        parsed_datas.insert(0, self.override_datas)

        self.__dict__["datas"] = tuple(parsed_datas)

    def __getattr__(self, key: str):
        for data in extract_callable_datas(self.datas):
            try:
                # Take a shallow copy of the object here, we don't want modifications
                # to host.data.<X> to stick, instead setting host.data.<Y> = is the
                # correct way to achieve this (see __setattr__).
                return copy(data[key])
            except KeyError:
                pass

        raise AttributeError(f"Host `{self.host}` has no data `{key}`")

    @override
    def __setattr__(self, key: str, value: Any):
        self.override_datas[key] = value

    @override
    def __str__(self):
        return str(self.datas)

    def get(self, key: str, default=None):
        return getattr(self, key, default)

    def dict(self):
        out = {}

        # Copy and reverse data objects (such that the first items override
        # the last, matching __getattr__ output).
        datas = list(self.datas)
        datas.reverse()

        for data in extract_callable_datas(datas):
            out.update(data)

        return out


class Host:
    """
    Represents a target host. Thin class that links up to facts and host/group
    data.
    """

    state: "State"
    connector_cls: type[BaseConnector]
    connector: BaseConnector
    connected: bool = False

    # Current context inside an @operation function (op gen stage)
    in_op: bool = False
    in_callback_op: bool = False
    current_op_hash: Optional[str] = None
    current_op_global_arguments: Optional["AllArguments"] = None

    # Current context inside a @deploy function which become part of the op data
    in_deploy: bool = False
    current_deploy_name: Optional[str] = None
    current_deploy_kwargs = None

    # @deploy decorator data is a bit different - we need to handle the case
    # where we're evaluating an operation at runtime (current_op_) but also
    # when ordering operations (current_) outside of an operation context.
    current_op_deploy_data: Optional[dict[str, Any]] = None
    current_deploy_data: Optional[dict[str, Any]] = None

    # Current context during operation execution
    executing_op_hash: Optional[str] = None
    nested_executing_op_hash: Optional[str] = None

    loop_position: list[int]

    # Arbitrary data dictionary connectors may use
    connector_data: dict[str, Any]

    def loop(self, iterable):
        self.loop_position.append(0)
        for i, item in enumerate(iterable):
            self.loop_position[-1] = i
            yield item
        self.loop_position.pop()

    def __init__(
        self,
        name: str,
        inventory: "Inventory",
        groups,
        connector_cls=None,
    ):
        if connector_cls is None:
            connector_cls = get_execution_connector("ssh")
        self.inventory = inventory
        self.groups = groups
        self.connector_cls = connector_cls
        self.name = name

        self.loop_position = []

        self.connector_data = {}

        # Append only list of operation hashes as called on this host, used to
        # generate a DAG to create the final operation order.
        self.op_hash_order: list[str] = []

        # Create the (waterfall data: override, host, group, global)
        self.data = HostData(
            self,
            lambda: inventory.get_override_data(),
            lambda: inventory.get_host_data(name),
            lambda: inventory.get_groups_data(groups),
            lambda: inventory.get_data(),
            # @deploy function data are default values, so come last
            self.get_deploy_data,
        )

    def init(self, state: "State") -> None:
        self.state = state
        self.connector = self.connector_cls(state, self)

        longest_name_len = max([len(host.name) for host in self.inventory])
        padding_diff = longest_name_len - len(self.name)
        self.print_prefix_padding = "".join(" " for _ in range(0, padding_diff))

    @override
    def __str__(self):
        return "{0}".format(self.name)

    @override
    def __repr__(self):
        return "Host({0})".format(self.name)

    @property
    def host_data(self):
        return self.inventory.get_host_data(self.name)

    @property
    def group_data(self):
        return self.inventory.get_groups_data(self.groups)

    @property
    def print_prefix(self) -> str:
        if self.nested_executing_op_hash:
            return "{0}[{1}] {2}{3} ".format(
                click.style(""),  # reset
                click.style(self.name, bold=True),
                click.style("nested", "blue"),
                self.print_prefix_padding,
            )

        return "{0}[{1}]{2} ".format(
            click.style(""),  # reset
            click.style(self.name, bold=True),
            self.print_prefix_padding,
        )

    def style_print_prefix(self, *args, **kwargs) -> str:
        return "{0}[{1}]{2} ".format(
            click.style(""),  # reset
            click.style(self.name, *args, **kwargs),
            self.print_prefix_padding,
        )

    def log(self, message: str, log_func: Callable[[str], Any] = logger.info) -> None:
        log_func(f"{self.print_prefix}{message}")

    def log_styled(
        self, message: str, log_func: Callable[[str], Any] = logger.info, **kwargs
    ) -> None:
        message_styled = click.style(message, **kwargs)
        self.log(message_styled, log_func=log_func)

    def get_deploy_data(self):
        return self.current_op_deploy_data or self.current_deploy_data or {}

    def noop(self, description: str) -> None:
        """
        Log a description for a noop operation.
        """

        handler = logger.info if self.state.print_noop_info else logger.debug
        handler("{0}noop: {1}".format(self.print_prefix, description))

    def when(self, condition: Callable[[], bool]):
        return self.deploy(
            "",
            cast("AllArguments", {"_if": [condition]}),
            {},
            in_deploy=False,
        )

    def arguments(self, **arguments: Unpack["AllArguments"]):
        return self.deploy("", arguments, {}, in_deploy=False)

    @contextmanager
    def deploy(
        self,
        name: str,
        kwargs: Optional["AllArguments"],
        data: Optional[dict],
        in_deploy: bool = True,
    ):
        """
        Wraps a group of operations as a deploy, this should not be used
        directly, instead use ``pyinfra.api.deploy.deploy``.
        """

        # Handle nested deploy names
        if self.current_deploy_name:
            name = "{0} | {1}".format(self.current_deploy_name, name)

        # Store the previous values
        old_in_deploy = self.in_deploy
        old_deploy_name = self.current_deploy_name
        old_deploy_kwargs = self.current_deploy_kwargs
        old_deploy_data = self.current_deploy_data
        self.in_deploy = in_deploy

        # Combine any old _ifs with the new ones
        if old_deploy_kwargs and kwargs:
            old_ifs = old_deploy_kwargs["_if"]
            new_ifs = kwargs["_if"]
            if old_ifs and new_ifs:
                kwargs["_if"] = old_ifs + new_ifs

        # Set the new values
        self.current_deploy_name = name
        self.current_deploy_kwargs = kwargs
        self.current_deploy_data = data
        logger.debug(
            "Starting deploy %s (args=%r, data=%r)",
            name,
            kwargs,
            data,
        )

        yield

        # Restore the previous values
        self.in_deploy = old_in_deploy
        self.current_deploy_name = old_deploy_name
        self.current_deploy_kwargs = old_deploy_kwargs
        self.current_deploy_data = old_deploy_data

        logger.debug(
            "Reset deploy to %s (args=%r, data=%r)",
            old_deploy_name,
            old_deploy_kwargs,
            old_deploy_data,
        )

    @memoize
    def _get_temp_directory(self):
        temp_directory = self.state.config.TEMP_DIR

        if temp_directory is None:
            # Unfortunate, but very hard to avoid, circular dependency, this method is memoized so
            # performance isn't a concern.
            from pyinfra.facts.server import TmpDir

            temp_directory = self.get_fact(TmpDir)

        if not temp_directory:
            temp_directory = self.state.config.DEFAULT_TEMP_DIR

        return temp_directory

    def get_temp_dir_config(self):
        return self.state.config.TEMP_DIR or self.state.config.DEFAULT_TEMP_DIR

    def get_temp_filename(
        self,
        hash_key: Optional[str] = None,
        hash_filename: bool = True,
        *,
        temp_directory: Optional[str] = None,
    ):
        """
        Generate a temporary filename for this deploy.
        """

        temp_directory = temp_directory or self._get_temp_directory()

        if not hash_key:
            hash_key = str(uuid4())

        if hash_filename:
            hash_key = sha1_hash(hash_key)

        return "{0}/pyinfra-{1}".format(temp_directory, hash_key)

    # Host facts
    #

    T = TypeVar("T")

    @overload
    def get_fact(self, name_or_cls: Type[FactBase[T]], *args, **kwargs) -> T: ...

    @overload
    def get_fact(self, name_or_cls: Type[ShortFactBase[T]], *args, **kwargs) -> T: ...

    def get_fact(self, name_or_cls, *args, **kwargs):
        """
        Get a fact for this host, reading from the cache if present.
        """
        return get_fact(self.state, self, name_or_cls, args=args, kwargs=kwargs)

    # Connector proxy
    #

    def _check_state(self) -> None:
        if not self.state:
            raise TypeError("Cannot call this function with no state!")

    def connect(self, reason=None, show_errors: bool = True, raise_exceptions: bool = False):
        """
        Connect to the host using it's configured connector.
        """

        self._check_state()
        if not self.connected:
            self.state.trigger_callbacks("host_before_connect", self)

            try:
                self.connector.connect()
            except ConnectError as e:
                if show_errors:
                    log_message = "{0}{1}".format(
                        self.print_prefix,
                        click.style(e.args[0], "red"),
                    )
                    logger.error(log_message)

                self.state.trigger_callbacks("host_connect_error", self, e)

                if raise_exceptions:
                    raise
            else:
                log_message = "{0}{1}".format(
                    self.print_prefix,
                    click.style("Connected", "green"),
                )
                if reason:
                    log_message = "{0}{1}".format(
                        log_message,
                        " ({0})".format(reason),
                    )

                logger.info(log_message)
                self.state.trigger_callbacks("host_connect", self)
                self.connected = True

    def disconnect(self) -> None:
        """
        Disconnect from the host using it's configured connector.
        """
        self._check_state()

        # Disconnect is an optional function for connectors if needed
        disconnect_func = getattr(self.connector, "disconnect", None)
        if disconnect_func:
            disconnect_func()

        # TODO: consider whether this should be here!
        remove_any_sudo_askpass_file(self)

        self.state.trigger_callbacks("host_disconnect", self)
        self.connected = False

    def run_shell_command(self, *args, **kwargs) -> tuple[bool, CommandOutput]:
        """
        Low level method to execute a shell command on the host via it's configured connector.
        """
        self._check_state()
        return self.connector.run_shell_command(*args, **kwargs)

    def put_file(self, *args, **kwargs) -> bool:
        """
        Low level method to upload a file to the host via it's configured connector.
        """
        self._check_state()
        return self.connector.put_file(*args, **kwargs)

    def get_file(self, *args, **kwargs) -> bool:
        """
        Low level method to download a file from the host via it's configured connector.
        """
        self._check_state()
        return self.connector.get_file(*args, **kwargs)

    # Rsync - optional connector specific ability

    def check_can_rsync(self) -> None:
        self._check_state()
        return self.connector.check_can_rsync()

    def rsync(self, *args, **kwargs) -> bool:
        self._check_state()
        return self.connector.rsync(*args, **kwargs)



from __future__ import annotations

import json
import os
from tempfile import mkstemp
from typing import TYPE_CHECKING

import click
from typing_extensions import TypedDict, Unpack, override

from pyinfra import local, logger
from pyinfra.api import QuoteString, StringCommand
from pyinfra.api.exceptions import ConnectError, InventoryError, PyinfraError
from pyinfra.api.util import get_file_io
from pyinfra.progress import progress_spinner

from .base import BaseConnector, DataMeta
from .local import LocalConnector
from .util import CommandOutput, extract_control_arguments, make_unix_command_for_host

if TYPE_CHECKING:
    from pyinfra.api.arguments import ConnectorArguments
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


class ConnectorData(TypedDict):
    docker_identifier: str
    docker_platform: str
    docker_architecture: str


connector_data_meta: dict[str, DataMeta] = {
    "docker_identifier": DataMeta("ID of container or image to start from"),
    "docker_platform": DataMeta("Platform to use for Docker image (e.g., linux/amd64)"),
    "docker_architecture": DataMeta("Architecture to use for Docker image (e.g., amd64, arm64)"),
}


class DockerConnector(BaseConnector):
    """
    The Docker connector allows you to use pyinfra to create new Docker images or modify running
    Docker containers.

    .. note::

        The Docker connector allows pyinfra to target Docker containers as inventory and is
        unrelated to the :doc:`../operations/docker` & :doc:`../facts/docker`.

    You can pass either an image name or existing container ID:

    + Image - will create a new container from the image, execute operations against it, save into \
        a new Docker image and remove the container
    + Existing container ID - will execute operations against the running container, leaving it \
        running

    .. code:: shell

        # A Docker base image must be provided
        pyinfra @docker/alpine:3.8 ...

        # pyinfra can run on multiple Docker images in parallel
        pyinfra @docker/alpine:3.8,@docker/ubuntu:bionic ...

        # Execute against a running container
        pyinfra @docker/2beb8c15a1b1 ...

    The Docker connector is great for testing pyinfra operations locally, rather than connecting to
    a remote host over SSH each time. This gives you a fast, local-first devloop to iterate on when
    writing deploys, operations or facts.
    """

    # enable the use of other docker cli compatible tools like podman
    docker_cmd = "docker"

    handles_execution = True

    data_cls = ConnectorData
    data_meta = connector_data_meta
    data: ConnectorData

    local: LocalConnector

    container_id: str
    no_stop: bool = False

    def __init__(self, state: "State", host: "Host"):
        super().__init__(state, host)
        self.local = LocalConnector(state, host)

    @override
    @staticmethod
    def make_names_data(name=None):
        if not name:
            raise InventoryError("No docker base ID provided!")

        yield (
            f"@docker/{name}",
            {"docker_identifier": name},
            ["@docker"],
        )

    # 2 helper functions
    def _find_start_docker_container(self, container_id) -> tuple[str, bool]:
        docker_info = local.shell(f"{self.docker_cmd} container inspect {container_id}")
        assert isinstance(docker_info, str)
        docker_info = json.loads(docker_info)[0]
        if docker_info["State"]["Running"] is False:
            logger.info(f"Starting stopped container: {container_id}")
            local.shell(f"{self.docker_cmd} container start {container_id}")
            return container_id, False
        return container_id, True

    def _start_docker_image(self, image_name):
        docker_cmd_parts = [
            self.docker_cmd,
            "run",
            "-d",
        ]

        if self.data.get("docker_platform"):
            docker_cmd_parts.extend(["--platform", self.data["docker_platform"]])
        if self.data.get("docker_architecture"):
            docker_cmd_parts.extend(["--arch", self.data["docker_architecture"]])

        docker_cmd_parts.extend(
            [
                image_name,
                "tail",
                "-f",
                "/dev/null",
            ]
        )

        try:
            return local.shell(
                " ".join(docker_cmd_parts),
                splitlines=True,
            )[-1]  # last line is the container ID
        except PyinfraError as e:
            raise ConnectError(e.args[0])

    @override
    def connect(self) -> None:
        self.local.connect()

        docker_identifier = self.data["docker_identifier"]
        with progress_spinner({f"prepare {self.docker_cmd} container"}):
            try:
                self.container_id, was_running = self._find_start_docker_container(
                    docker_identifier
                )
                if was_running:
                    self.no_stop = True
            except PyinfraError:
                self.container_id = self._start_docker_image(docker_identifier)

    @override
    def disconnect(self) -> None:
        container_id = self.container_id

        if self.no_stop:
            logger.info(
                "{0}{1} build complete, container left running: {2}".format(
                    self.host.print_prefix,
                    self.docker_cmd,
                    click.style(container_id, bold=True),
                ),
            )
            return

        with progress_spinner({f"{self.docker_cmd} commit"}):
            image_id = local.shell(f"{self.docker_cmd} commit {container_id}", splitlines=True)[-1][
                7:19
            ]  # last line is the image ID, get sha256:[XXXXXXXXXX]...

        with progress_spinner({f"{self.docker_cmd} rm"}):
            local.shell(
                f"{self.docker_cmd} rm -f {container_id}",
            )

        logger.info(
            "{0}{1} build complete, image ID: {2}".format(
                self.host.print_prefix,
                self.docker_cmd,
                click.style(image_id, bold=True),
            ),
        )

    @override
    def run_shell_command(
        self,
        command: StringCommand,
        print_output: bool = False,
        print_input: bool = False,
        **arguments: Unpack["ConnectorArguments"],
    ) -> tuple[bool, CommandOutput]:
        local_arguments = extract_control_arguments(arguments)

        container_id = self.container_id

        command = make_unix_command_for_host(self.state, self.host, command, **arguments)
        command = StringCommand(QuoteString(command))

        docker_flags = "-it" if local_arguments.get("_get_pty") else "-i"
        docker_command = StringCommand(
            self.docker_cmd,
            "exec",
            docker_flags,
            container_id,
            "sh",
            "-c",
            command,
        )

        return self.local.run_shell_command(
            docker_command,
            print_output=print_output,
            print_input=print_input,
            **local_arguments,
        )

    @override
    def put_file(
        self,
        filename_or_io,
        remote_filename,
        remote_temp_filename=None,  # ignored
        print_output=False,
        print_input=False,
        **kwargs,  # ignored (sudo/etc)
    ) -> bool:
        """
        Upload a file/IO object to the target container by copying it to a
        temporary location and then uploading it into the container using ``docker cp``.
        """

        fd, temp_filename = mkstemp()

        try:
            # Load our file or IO object and write it to the temporary file
            with get_file_io(filename_or_io) as file_io:
                with open(temp_filename, "wb") as temp_f:
                    data = file_io.read()

                    if isinstance(data, str):
                        data = data.encode()

                    temp_f.write(data)

            docker_command = StringCommand(
                self.docker_cmd,
                "cp",
                temp_filename,
                f"{self.container_id}:{remote_filename}",
            )

            status, output = self.local.run_shell_command(
                docker_command,
                print_output=print_output,
                print_input=print_input,
            )
        finally:
            os.close(fd)
            os.remove(temp_filename)

        if not status:
            raise IOError(output.stderr)

        if print_output:
            click.echo(
                "{0}file uploaded to container: {1}".format(
                    self.host.print_prefix,
                    remote_filename,
                ),
                err=True,
            )

        return status

    @override
    def get_file(
        self,
        remote_filename,
        filename_or_io,
        remote_temp_filename=None,  # ignored
        print_output=False,
        print_input=False,
        **kwargs,  # ignored (sudo/etc)
    ) -> bool:
        """
        Download a file from the target container by copying it to a temporary
        location and then reading that into our final file/IO object.
        """

        fd, temp_filename = mkstemp()

        try:
            docker_command = StringCommand(
                self.docker_cmd,
                "cp",
                f"{self.container_id}:{remote_filename}",
                temp_filename,
            )

            status, output = self.local.run_shell_command(
                docker_command,
                print_output=print_output,
                print_input=print_input,
            )

            # Load the temporary file and write it to our file or IO object
            with open(temp_filename, "rb") as temp_f:
                with get_file_io(filename_or_io, "wb") as file_io:
                    data = temp_f.read()
                    file_io.write(data)
        finally:
            os.close(fd)
            os.remove(temp_filename)

        if not status:
            raise IOError(output.stderr)

        if print_output:
            click.echo(
                "{0}file downloaded from container: {1}".format(
                    self.host.print_prefix,
                    remote_filename,
                ),
                err=True,
            )

        return status


class PodmanConnector(DockerConnector):
    """
    The Podman connector allows you to use pyinfra to create new Podman images or modify running
    Podman containers.

    .. note::

        The Podman connector allows pyinfra to target Podman containers as inventory and is
        unrelated to the :doc:`../operations/docker` & :doc:`../facts/docker`.

    You can pass either an image name or existing container ID:

    + Image - will create a new container from the image, execute operations against it, save into \
        a new Podman image and remove the container
    + Existing container ID - will execute operations against the running container, leaving it \
        running

    .. code:: shell

        # A Podman base image must be provided
        pyinfra @podman/alpine:3.8 ...

        # pyinfra can run on multiple Docker images in parallel
        pyinfra @podman/alpine:3.8,@podman/ubuntu:bionic ...

        # Execute against a running container
        pyinfra @podman/2beb8c15a1b1 ...

    The Podman connector is great for testing pyinfra operations locally, rather than connecting to
    a remote host over SSH each time. This gives you a fast, local-first devloop to iterate on when
    writing deploys, operations or facts.
    """

    docker_cmd = "podman"

    @override
    @staticmethod
    def make_names_data(name=None):
        if not name:
            raise InventoryError("No podman base ID provided!")

        yield (
            f"@podman/{name}",
            {"docker_identifier": name},
            ["@podman"],
        )

    # Duplicate function definition to swap the docstring.
    @override
    def put_file(
        self,
        filename_or_io,
        remote_filename,
        remote_temp_filename=None,  # ignored
        print_output=False,
        print_input=False,
        **kwargs,  # ignored (sudo/etc)
    ) -> bool:
        """
        Upload a file/IO object to the target container by copying it to a
        temporary location and then uploading it into the container using ``podman cp``.
        """
        return super().put_file(
            filename_or_io,
            remote_filename,
            remote_temp_filename,  # ignored
            print_output,
            print_input,
            **kwargs,  # ignored (sudo/etc)
        )



"""
Docker based integration tests.
"""

import pytest
import testinfra
import testinfra.host


@pytest.mark.end_to_end
@pytest.mark.end_to_end_docker
def test_int_docker_install_package_ubuntu(helpers):
    def check(host: testinfra.host.Host):
        assert host.package("iftop").is_installed

    helpers.run_container_test_host(
        "ubuntu:22.04",
        "apt.packages packages=iftop update=true",
        check,
    )


@pytest.mark.end_to_end
@pytest.mark.end_to_end_docker
def test_int_docker_file(helpers):
    def check(host: testinfra.host.Host):
        file = host.file("/testfile")
        assert file.is_file
        assert file.user == "nobody"
        assert file.group == "root"
        assert file.mode == 0o755

    helpers.run_container_test_host(
        "ubuntu:22.04",
        "files.file path=/testfile mode=755 user=nobody",
        check,
    )


@pytest.mark.end_to_end
@pytest.mark.end_to_end_docker
def test_int_docker_put_file(helpers):
    # TODO: why does importing this at top level break things?
    from pyinfra.api.util import get_file_sha256

    with open("README.md", "r") as f:
        expected_sum = get_file_sha256(f)

    def check(host: testinfra.host.Host):
        file = host.file("/README.md")
        assert file.is_file
        assert file.sha256sum == expected_sum

    helpers.run_container_test_host(
        "ubuntu:22.04",
        "files.put src=README.md dest=README.md mode=755 user=nobody",
        check,
    )
